###################################
# Parse a QMASM source file       #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import copy
import os
import qmasm
import re
import random
import string
import sys
from collections import defaultdict
from qmasm.assertions import AssertParser
from qmasm.utils import RemainingNextException

# Define a function that aborts the program, reporting an invalid
# input line as part of the error message.
def error_in_line(filename, lineno, str):
    sys.stderr.write('%s:%d: error: %s\n' % (filename, lineno, str))
    sys.exit(1)

class Environment(object):
    "Maintain a variable environment as a stack of scopes."

    toks_re = re.compile(r'([^-+*/%&\|^~()<=>#,\s\[:\]]+)')  # Regex to split a symbol into tokens (cf. qmasm.ident_re with square brackets and colons added)

    def __init__(self):
        self.stack = [{}]
        self.self_copy = None   # Regenerate a copy of self if equal to None

    def __getitem__(self, key):
        "Search each scope in turn for the given key."
        for i in range(len(self.stack) - 1, -1, -1):
            try:
                return self.stack[i][key]
            except KeyError:
                pass
        raise KeyError(key)

    def __setitem__(self, key, val):
        self.stack[-1][key] = val
        self.self_copy = None

    def push(self):
        "Push a new scope on the environment stack."
        self.stack.append({})
        self.self_copy = None

    def pop(self):
        "Discard the top of the environment stack."
        self.stack.pop()
        self.self_copy = None

    def keys(self):
        "Return all keys in all scopes."
        d = set()
        for s in self.stack:
            d.update(s.keys())
        return list(d)

    def __copy__(self):
        "Return an independent copy of the environment stack."
        if self.self_copy == None:
            self.self_copy = copy.deepcopy(self)
        return self.self_copy

    def sub_syms(self, sym):
        """Substitute values for variables encountered in a given symbol name
        or list of symbol names."""
        if isinstance(sym, (list,)):
            # Recursively process each list element in turn.
            return [self.sub_syms(s) for s in sym]
        elif isinstance(sym, (str,)):
            # Process a single word.
            toks = self.toks_re.split(sym)
            for i in range(len(toks)):
                try:
                    toks[i] = str(self[toks[i]])
                except KeyError:
                    pass
            return "".join(toks)
        else:
            # Non-strings are returned unmodified.
            return sym

# I'm too lazy to write another parser so I'll simply define an
# alternative entry point to the assertion parser.
class ExprParser(AssertParser):
    "Parse an arithmetic expression."

    def parse(self, filename, lineno, s):
        "Parse an arithmetic expression into an AST"
        self.tokens = self.lex(s)
        self.tokidx = -1
        self.advance()
        try:
            ast = self.expression()
            if self.sym[0] != "EOF":
                raise self.ParseError('Parse error at "%s"' % self.sym[1])
        except self.ParseError as e:
            sys.stderr.write('%s:%d: error: %s in "%s"\n' % (filename, lineno, e, s))
            sys.exit(1)
        return ast

# I'm too lazy to write another parser so I'll simply define an
# alternative entry point to the assertion parser.
class RelationParser(AssertParser):
    "Parse a relational expression."

    def parse(self, filename, lineno, s):
        "Parse a relational expression into an AST"
        self.tokens = self.lex(s)
        self.tokidx = -1
        self.advance()
        try:
            ast = self.conjunction()
            if self.sym[0] != "EOF":
                raise self.ParseError('Parse error at "%s"' % self.sym[1])
        except self.ParseError as e:
            sys.stderr.write('%s:%d: error: %s in "%s"\n' % (filename, lineno, e, s))
            sys.exit(1)
        return ast

class LoopIterator(object):
    '''Iterate over arithmetic and geometric integer sequences.  It is
    assumed that the penultimate element of the input list is the string
    "...".'''

    def __init__(self, filename, lineno, rhs):
        # Prepare the iterations we intend to perform.
        self.first_val = rhs[0]
        last_val = rhs[-1]
        if len(rhs) == 3:
            # "<x_0> ... <x_n>" indicates an arithmetic progression with a
            # delta of +/- 1.
            if self.first_val <= last_val:
                self.increment = lambda x: x + 1
            else:
                self.increment = lambda x: x - 1
        else:
            # For "<x_0> <x_1> <x_2> ... <x_n>", compute the progression
            # type and increment.
            adeltas = set([rhs[i + 1] - rhs[i] for i in range(len(rhs) - 3)])
            if len(adeltas) == 1:
                # Arithmetic progression
                delta = list(adeltas)[0]
                self.increment = lambda x: x + delta
            else:
                # Geometric progression
                try:
                    mdeltas = set([rhs[i + 1] // rhs[i] for i in range(len(rhs) - 3)])
                except ZeroDivisionError:
                    mdeltas = set(["multiple", "values"])  # Force the next test to fail.
                if len(mdeltas) != 1:
                    error_in_line(filename, lineno,
                                  'Failed to interpret "%s, %d" as either an arithmetic or geometric progression' %
                                  (", ".join([str(r) for r in rhs[:-2]]), last_val))
                delta = list(mdeltas)[0]
                if delta == 0:
                    error_in_line(filename, lineno, "Decreasing geometric progressions are not currently supported")
                self.increment = lambda x: x * delta

        # Determine when to stop.
        second_val = self.increment(self.first_val)
        if self.first_val < second_val:
            self.finished = lambda x: x > last_val
        else:
            self.finished = lambda x: x < last_val

    def __iter__(self):
        self.next_val = self.first_val
        return self

    def __next__(self):
        if self.finished(self.next_val):
            raise StopIteration
        head = self.next_val
        self.next_val = self.increment(head)
        return head

    def next(self):
        # Python 2 wrapper for __next__
        return self.__next__()

class Statement(object):
    "One statement in a QMASM source file."

    def __init__(self, qmasm, filename, lineno):
        self.qmasm = qmasm
        self.filename = filename
        self.lineno = lineno

    def error_in_line(self, msg):
        if self.lineno == None:
            self.qmasm.abend(msg)
        else:
            sys.stderr.write('%s:%d: error: %s\n' % (self.filename, self.lineno, msg))
        sys.exit(1)

    def validate_ident(self, ident):
        """Complain if an identifier uses invalid symbols.  Otherwise, return
        the argument unmodified."""
        match = self.qmasm.ident_re.match(ident)
        if match == None or match.group(0) != ident:
            self.error_in_line('Invalid identifier "%s"' % ident)
        return ident

class Weight(Statement):
    "Represent a point weight on a qubit."
    def __init__(self, qmasm, filename, lineno, sym, weight):
        super(Weight, self).__init__(qmasm, filename, lineno)
        self.sym = self.validate_ident(sym)
        self.weight = weight

    def as_str(self, prefix=""):
        return "%s%s %s" % (prefix, self.sym, self.weight)

    def update_qmi(self, prefix, next_prefix, problem):
        num = self.qmasm.symbol_to_number(prefix + self.sym, prefix, next_prefix)
        problem.weights[num] += self.weight

class Chain(Statement):
    "Chain between qubits."
    def __init__(self, qmasm, filename, lineno, sym1, sym2):
        super(Chain, self).__init__(qmasm, filename, lineno)
        self.sym1 = self.validate_ident(sym1)
        self.sym2 = self.validate_ident(sym2)

    def as_str(self, prefix=""):
        return "%s%s = %s%s" % (prefix, self.sym1, prefix, self.sym2)

    def update_qmi(self, prefix, next_prefix, problem):
        num1 = self.qmasm.symbol_to_number(prefix + self.sym1, prefix, next_prefix)
        num2 = self.qmasm.symbol_to_number(prefix + self.sym2, prefix, next_prefix)
        if num1 == num2:
            self.error_in_line("A chain cannot connect a spin to itself")
        elif num1 > num2:
            num1, num2 = num2, num1
        problem.chains.add((num1, num2))
        sym1 = self.qmasm.apply_prefix(prefix + self.sym1, None, next_prefix)
        sym2 = self.qmasm.apply_prefix(prefix + self.sym2, None, next_prefix)
        problem.pending_asserts.append((sym1, "=", sym2))

class AntiChain(Statement):
    "AntiChain between qubits."
    def __init__(self, qmasm, filename, lineno, sym1, sym2):
        super(AntiChain, self).__init__(qmasm, filename, lineno)
        self.sym1 = self.validate_ident(sym1)
        self.sym2 = self.validate_ident(sym2)

    def as_str(self, prefix=""):
        return "%s%s /= %s%s" % (prefix, self.sym1, prefix, self.sym2)

    def update_qmi(self, prefix, next_prefix, problem):
        num1 = self.qmasm.symbol_to_number(prefix + self.sym1, prefix, next_prefix)
        num2 = self.qmasm.symbol_to_number(prefix + self.sym2, prefix, next_prefix)
        if num1 == num2:
            self.error_in_line("An anti-chain cannot connect a spin to itself")
        elif num1 > num2:
            num1, num2 = num2, num1
        problem.antichains.add((num1, num2))
        sym1 = self.qmasm.apply_prefix(prefix + self.sym1, None, next_prefix)
        sym2 = self.qmasm.apply_prefix(prefix + self.sym2, None, next_prefix)
        problem.pending_asserts.append((sym1, "/=", sym2))

class Pin(Statement):
    "Pinning of a qubit to true or false."
    def __init__(self, qmasm, filename, lineno, sym, goal):
        super(Pin, self).__init__(qmasm, filename, lineno)
        self.sym = self.validate_ident(sym)
        self.goal = goal

    def as_str(self, prefix=""):
        return "%s%s := %s" % (prefix, self.sym, self.goal)

    def update_qmi(self, prefix, next_prefix, problem):
        num = self.qmasm.symbol_to_number(prefix + self.sym, prefix, next_prefix)
        problem.pinned.append((num, self.goal))
        sym = self.qmasm.apply_prefix(prefix + self.sym, None, next_prefix)
        problem.pending_asserts.append((sym, "=", str(int(self.goal))))

class Alias(Statement):
    "Alias of one symbol to another."
    def __init__(self, qmasm, filename, lineno, sym1, sym2):
        super(Alias, self).__init__(qmasm, filename, lineno)
        self.sym1 = self.validate_ident(sym1)
        self.sym2 = self.validate_ident(sym2)

    def as_str(self, prefix=""):
        return "%s%s <-> %s%s" % (prefix, self.sym1, prefix, self.sym2)

    def update_qmi(self, prefix, next_prefix, problem):
        sym1 = prefix + self.sym1
        sym2 = prefix + self.sym2
        if next_prefix != None:
            sym1 = sym1.replace(prefix + "!next.", next_prefix)
            sym2 = sym2.replace(prefix + "!next.", next_prefix)
        self.qmasm.sym_map.alias(sym1, sym2)

class Rename(Statement):
    "Rename one set of symbols to another."
    def __init__(self, qmasm, filename, lineno, syms1, syms2):
        super(Rename, self).__init__(qmasm, filename, lineno)
        self.syms1 = [self.validate_ident(s) for s in syms1]
        self.syms2 = [self.validate_ident(s) for s in syms2]

    def as_str(self, prefix=""):
        return " ".join([prefix + s for s in self.syms1]) + " -> " + " ".join([prefix + s for s in self.syms2])

    def update_qmi(self, prefix, next_prefix, problem):
        # Update the symbol map.
        syms1 = [prefix + s for s in self.syms1]
        syms2 = [prefix + s for s in self.syms2]
        if next_prefix != None:
            syms1 = [s.replace(prefix + "!next.", next_prefix) for s in syms1]
            syms2 = [s.replace(prefix + "!next.", next_prefix) for s in syms2]
        self.qmasm.sym_map.replace_all(syms1, syms2)
        sym2sym = dict(zip(syms1, syms2))

        # Update all weights.
        weights = {}
        for q, wt in problem.weights.items():
            try:
                q = sym2sym[q]
            except KeyError:
                pass
            weights[q] = wt
        problem.weights = defaultdict(lambda: 0.0, weights)

        # Update all strengths.
        strengths = {}
        for (q1, q2), wt in problem.strengths.items():
            try:
                q1 = sym2sym[q1]
            except KeyError:
                pass
            try:
                q2 = sym2sym[q2]
            except KeyError:
                pass
            strengths[(q1, q2)] = wt
        problem.strengths = defaultdict(lambda: 0.0, strengths)

        # Update all chains.
        chains = set()
        for (q1, q2) in problem.chains:
            try:
                q1 = sym2sym[q1]
            except KeyError:
                pass
            try:
                q2 = sym2sym[q2]
            except KeyError:
                pass
            chains.add((q1, q2))
        problem.chains = chains

        # Update all anti-chains.
        antichains = set()
        for (q1, q2) in problem.antichains:
            try:
                q1 = sym2sym[q1]
            except KeyError:
                pass
            try:
                q2 = sym2sym[q2]
            except KeyError:
                pass
            antichains.add((q1, q2))
        problem.antichains = antichains

        # Update all assertions.  These need to go through an intermediary in
        # case we rename both X to Y and Y to X.
        renames = []
        for s1, s2 in sym2sym.items():
            dummy_sym = "".join([random.choice(string.ascii_lowercase) for i in range(5)])
            dummy_sym += " "   # Can't currently appear in a symbol name.
            dummy_sym += "".join([random.choice(string.ascii_lowercase) for i in range(5)])
            renames.append((s1, dummy_sym, s2))
        for s1, dummy_sym, s2 in renames:
            for ast in problem.assertions:
                ast.replace_ident(s1, dummy_sym)
        for s1, dummy_sym, s2 in renames:
            for ast in problem.assertions:
                ast.replace_ident(dummy_sym, s2)

        # Update all pending assertions.
        pending_asserts = []
        for s1, op, s2 in problem.pending_asserts:
            try:
                s1 = sym2sym[s1]
            except KeyError:
                pass
            try:
                s2 = sym2sym[s2]
            except KeyError:
                pass
            pending_asserts.append((s1, op, s2))
        problem.pending_asserts = pending_asserts

class Strength(Statement):
    "Coupler strength between two qubits."
    def __init__(self, qmasm, filename, lineno, sym1, sym2, strength):
        super(Strength, self).__init__(qmasm, filename, lineno)
        self.sym1 = self.validate_ident(sym1)
        self.sym2 = self.validate_ident(sym2)
        self.strength = strength

    def as_str(self, prefix=""):
        return "%s%s %s%s %s" % (prefix, self.sym1, prefix, self.sym2, self.strength)

    def update_qmi(self, prefix, next_prefix, problem):
        num1 = self.qmasm.symbol_to_number(prefix + self.sym1, prefix, next_prefix)
        num2 = self.qmasm.symbol_to_number(prefix + self.sym2, prefix, next_prefix)
        if num1 == num2:
            self.error_in_line("A coupler cannot connect a spin to itself")
        elif num1 > num2:
            num1, num2 = num2, num1
        problem.strengths[(num1, num2)] += self.strength

class Assert(Statement):
    "Instantiation of a run-time assertion."

    def __init__(self, qmasm, filename, lineno, expr):
        super(Assert, self).__init__(qmasm, filename, lineno)
        self.parser = AssertParser(qmasm)
        self.expr = expr
        self.ast = self.parser.parse(expr)
        self.ast.compile()

    def as_str(self, prefix=""):
        if prefix == "":
            ast = self.ast
        else:
            ast = copy.deepcopy(self.ast)
            ast.prefix_identifiers(prefix, None)
            ast.compile()
        return "!assert " + str(ast)

    def update_qmi(self, prefix, next_prefix, problem):
        if prefix == "":
            ast = self.ast
        else:
            ast = copy.deepcopy(self.ast)
            ast.prefix_identifiers(prefix, next_prefix)
            ast.compile()
        problem.assertions.append(ast)

class MacroUse(Statement):
    "Instantiation of a macro definition."
    def __init__(self, qmasm, filename, lineno, name, body, prefixes):
        super(MacroUse, self).__init__(qmasm, filename, lineno)
        self.name = self.validate_ident(name)
        self.body = body
        self.prefixes = [self.validate_ident(p) + "." for p in prefixes]

    def as_str(self, prefix=""):
        stmt_strs = []
        nprefixes = len(self.prefixes)
        if nprefixes == 0:
            # No prefixes -- display the macro body in the current scope (i.e.,
            # using the given prefix).
            for stmt in self.body:
                sstr = stmt.as_str(prefix)
                if "!next." not in sstr:
                    stmt_strs.append(sstr)
        else:
            # At least one prefix -- disaply the macro body in a new scope
            # (i.e., by augmenting the given prefix with each new prefix in
            # turn).
            for p in range(nprefixes):
                pfx = self.prefixes[p]
                for stmt in self.body:
                    sstr = stmt.as_str(prefix + pfx)
                    if "!next." in sstr:
                        if p == nprefixes - 1:
                            # Drop statements that use "!next." if there's
                            # no next prefix.
                            continue
                        next_pfx = self.prefixes[p + 1]
                        sstr = sstr.replace(prefix + pfx + "!next.", prefix + next_pfx)
                    stmt_strs.append(sstr)
        return "\n".join(stmt_strs)

    def update_qmi(self, prefix, next_prefix, problem):
        nprefixes = len(self.prefixes)
        if nprefixes == 0:
            # No prefixes -- import the macro body into the current scope
            # (i.e., using the given prefix).
            for stmt in self.body:
                try:
                    stmt.update_qmi(prefix, None, problem)
                except self.qmasm.utils.RemainingNextException:
                    pass
        else:
            # At least one prefix -- import the macro body into a new scope
            # (i.e., by augmenting the given prefix with each new prefix in
            # turn).
            for p in range(nprefixes):
                pfx = prefix + self.prefixes[p]
                if p == nprefixes - 1:
                    next_pfx = None
                else:
                    next_pfx = prefix + self.prefixes[p + 1]
                for stmt in self.body:
                    try:
                        stmt.update_qmi(pfx, next_pfx, problem)
                    except RemainingNextException:
                        pass

class FileParser(object):
    "Parse a QMASM file."

    def __init__(self, qmasm):
        self.qmasm = qmasm      # Reference to the object we're mixed into
        self.macros = {}        # Map from a macro name to a list of Statement objects
        self.current_macro = (None, [])   # Macro currently being defined (name and statements)
        self.target = qmasm.program   # Reference to either the program or the current macro
        self.env = Environment()      # Stack of maps from compile-time variable names to values
        self.expr_parser = ExprParser(qmasm)     # Expression parser
        self.rel_parser = RelationParser(qmasm)  # Relation parser

        # Establish a mapping from a first-field directive to a parsing function.
        self.dir_to_func = {
            "!include":     self.parse_line_include,
            "!assert":      self.parse_line_assert,
            "!let":         self.parse_line_let,
            "!begin_macro": self.parse_line_begin_macro,
            "!end_macro":   self.parse_line_end_macro,
            "!use_macro":   self.parse_line_use_macro,
            "!alias":       self.parse_line_sym_alias
        }

    def is_float(self, str):
        "Return True if a string can be treated as a float."
        try:
            float(str)
            return True
        except ValueError:
            return False

    def split_line_into_fields(self, line):
        "Discard comments, then split the line into fields."
        try:
            line = line[:line.index("#")]
        except ValueError:
            pass
        return line.split()

    def find_file_in_path(self, pathnames, filename):
        "Search a list of directories for a file."
        for pname in pathnames:
            fname = os.path.join(pname, filename)
            if os.path.exists(fname):
                return fname
            fname_qmasm = fname + ".qmasm"
            if os.path.exists(fname_qmasm):
                return fname_qmasm
        return None

    def parse_line_include(self, filename, lineno, fields):
        "Parse an !include directive."
        # "!include" "<filename>" -- process a named auxiliary file.
        if len(fields) != 2:
            error_in_line(filename, lineno, "Expected a filename to follow !include")
        incname = " ".join(fields[1:])
        if len(incname) >= 2 and incname[0] == "<" and incname[-1] == ">":
            # Search QMASMPATH for the filename.
            incname = incname[1:-1]
            try:
                qmasmpath = os.environ["QMASMPATH"].split(":")
                qmasmpath.append(".")
            except KeyError:
                qmasmpath = ["."]
            found_incname = self.find_file_in_path(qmasmpath, incname)
            if found_incname != None:
                incname = found_incname
        elif len(incname) >= 2 and incname[0] == '"' and incname[-1] == '"':
            # Search only the current directory for the filename.
            incname = incname[1:-1]
            found_incname = self.find_file_in_path(["."], incname)
            if found_incname != None:
                incname = found_incname
        else:
            error_in_line(filename, lineno, 'Failed to parse "%s"' % (" ".join(fields)))
        try:
            incfile = open(incname)
        except IOError:
            error_in_line(filename, lineno, 'Failed to open %s for input' % incname)
        self.process_file(incname, incfile)
        incfile.close()

    def parse_line_assert(self, filename, lineno, fields):
        "Parse an !assert directive."
        # "!assert" <expr> -- assert a property that must be true at run time.
        if len(fields) < 2:
            error_in_line(filename, lineno, "Expected an expression to follow !assert")
        self.target.append(Assert(self.qmasm, filename, lineno, " ".join(self.env.sub_syms(fields[1:]))))

    def parse_line_let(self, filename, lineno, fields):
        "Parse a !let directive."
        # "!let" <name> := <expr> -- evaluate <expr> and assign the result to <name>.
        if len(fields) < 4 or fields[2] != ":=":
            error_in_line(filename, lineno, 'Expected a variable name, ":=", and an expression to follow !let')
        lhs = fields[1]
        if len(fields) == 4:
            # Handle the case of "!let" <name> := <symbol>.
            is_sym = self.env.toks_re.match(fields[3])
            if is_sym != None and is_sym.group(0) == fields[3]:
                self.env[lhs] = self.env.sub_syms(fields[3])
                return
        ast = self.expr_parser.parse(filename, lineno, " ".join(self.env.sub_syms(fields[3:])))
        ast.compile()
        rhs = ast.evaluate(dict(self.env))
        self.env[lhs] = rhs

    def parse_line_begin_macro(self, filename, lineno, fields):
        "Parse a !begin_macro directive."
        # "!begin_macro" <name> -- begin a macro definition.
        if len(fields) != 2:
            error_in_line(filename, lineno, "Expected a macro name to follow !begin_macro")
        name = fields[1]
        if name in self.macros:
            error_in_line(self, filename, lineno, "Macro %s is multiply defined" % name)
        if self.current_macro[0] != None:
            error_in_line(filename, lineno, "Nested macros are not supported")
        self.current_macro = (name, [])
        self.target = self.current_macro[1]
        self.env.push()

    def parse_line_end_macro(self, filename, lineno, fields):
        "Parse an !end_macro directive."
        # "!end_macro" <name> -- end a macro definition.
        if len(fields) != 2:
            error_in_line(filename, lineno, "Expected a macro name to follow !end_macro")
        name = fields[1]
        if self.current_macro[0] == None:
            error_in_line(filename, lineno, "Ended macro %s with no corresponding begin" % name)
        if self.current_macro[0] != name:
            error_in_line(filename, lineno, "Ended macro %s after beginning macro %s" % (name, self.current_macro[0]))
        self.macros[name] = self.current_macro[1]
        self.target = self.qmasm.program
        self.current_macro = (None, [])
        self.env.pop()

    def parse_line_weight(self, filename, lineno, fields):
        "Parse a qubit weight."
        # <symbol> <weight> -- increment a symbol's point weight.
        if len(fields) != 2:
            error_in_line(filename, lineno, "Internal error in parse_line_weight")
        try:
            val = float(self.env.sub_syms(fields[1]))
        except ValueError:
            error_in_line(filename, lineno, 'Failed to parse "%s %s" as a symbol followed by a numerical weight' % (fields[0], fields[1]))
        self.target.append(Weight(self.qmasm, filename, lineno, self.env.sub_syms(fields[0]), val))

    def parse_line_chain(self, filename, lineno, fields):
        "Parse a qubit chain."
        # <symbol_1> = <symbol_2> -- create a chain between <symbol_1>
        # and <symbol_2>.
        if len(fields) != 3 or fields[1] != "=":
            error_in_line(filename, lineno, "Internal error in parse_line_chain")
        code = "%s = %s" % (self.env.sub_syms(fields[0]), self.env.sub_syms(fields[2]))
        self.target.extend(self.process_chain(filename, lineno, code))

    def parse_line_antichain(self, filename, lineno, fields):
        "Parse a qubit anti-chain."
        # <symbol_1> /= <symbol_2> -- create an anti-chain between <symbol_1>
        # and <symbol_2>.
        if len(fields) != 3 or fields[1] != "/=":
            error_in_line(filename, lineno, "Internal error in parse_line_antichain")
        code = "%s /= %s" % (self.env.sub_syms(fields[0]), self.env.sub_syms(fields[2]))
        self.target.extend(self.process_antichain(filename, lineno, code))

    def parse_line_pin(self, filename, lineno, fields):
        "Parse a qubit pin."
        # <symbol> := <value> -- force symbol <symbol> to have value <value>.
        if len(fields) != 3 or fields[1] != ":=":
            error_in_line(filename, lineno, "Internal error in parse_line_pin")
        code = "%s := %s" % (self.env.sub_syms(fields[0]), self.env.sub_syms(fields[2]))
        self.target.extend(self.process_pin(filename, lineno, code))

    def parse_line_alias(self, filename, lineno, fields):
        "Parse a qubit alias."
        # <symbol_1> <-> <symbol_2> -- make <symbol_1> an alias of <symbol_2>.
        if len(fields) != 3 or fields[1] != "<->":
            error_in_line(filename, lineno, "Internal error in parse_line_alias")
        code = "%s <-> %s" % (self.env.sub_syms(fields[0]), self.env.sub_syms(fields[2]))
        self.target.extend(self.process_alias(filename, lineno, code))

    def parse_line_rename(self, filename, lineno, fields):
        "Parse a qubit rename."
        # <symbol_1> ... -> <symbol_2> ... -- make <symbol_1> an alias of <symbol_2>.
        if len(fields) < 3 or len(fields)%2 == 0:
            error_in_line(filename, lineno, 'Failed to parse "%s" as a symbol rename' % (" ".join(fields)))

        # Split the fields into a left-hand side and a right-hand side.
        pin_parser = PinParser()
        tokens = []
        num_arrows = 0
        for sym in fields:
            if sym == "->":
                tokens.append(sym)
                num_arrows += 1
            else:
                sym_list = pin_parser.parse_lhs(self.env.sub_syms(sym))
                tokens.extend(sym_list)
        lhs = tokens[:len(tokens)//2]
        rhs = tokens[len(tokens)//2:]
        if num_arrows != 1 or rhs[0] != "->":
            error_in_line(filename, lineno, 'Failed to parse "%s" as a symbol rename' % (" ".join(fields)))
        rhs = rhs[1:]  # Drop the "->".
        self.target.append(Rename(self.qmasm, filename, lineno, lhs, rhs))

    def parse_line_strength(self, filename, lineno, fields):
        "Parse a coupler strength."
        # <symbol_1> <symbol_2> <strength> -- increment a coupler strength.
        if len(fields) != 3:
            error_in_line(filename, lineno, "Internal error in parse_line_strength")
        try:
            strength = float(self.env.sub_syms(fields[2]))
        except ValueError:
            error_in_line(filename, lineno, 'Failed to parse "%s" as a number' % fields[2])
        self.target.append(Strength(self.qmasm, filename, lineno, self.env.sub_syms(fields[0]), self.env.sub_syms(fields[1]), strength))

    def parse_line_use_macro(self, filename, lineno, fields):
        "Parse a !use_macro directive."
        # "!use_macro" <macro_name> [<instance_name> ...] -- instantiate a
        # macro using <instance_name> as each variable's prefix.
        if len(fields) < 2:
            error_in_line(filename, lineno, "Expected a macro name to follow !use_macro")
        name = self.env.sub_syms(fields[1])
        prefixes = [self.env.sub_syms(p) for p in fields[2:]]
        try:
            self.target.append(MacroUse(self.qmasm, filename, lineno, name, self.macros[name], prefixes))
        except KeyError:
            error_in_line(filename, lineno, "Unknown macro %s" % name)

    def parse_line_sym_alias(self, filename, lineno, fields):
        "Parse an !alias directive."
        sys.stderr.write('%s:%d: warning: !alias is deprecated; use "!let %s := %s" instead\n' % (filename, lineno, fields[1], fields[2]))
        if len(fields) != 3:
            error_in_line(filename, lineno, "Expected a symbol name and replacement to follow !alias")
        self.env[fields[1]] = self.env.sub_syms(fields[2])

    def process_if(self, filename, lineno, fields, all_lines):
        """Parse and process an !if directive.  Recursively parse the remaining
        file contents."""
        if len(fields) < 2:
            error_in_line(filename, lineno, "Expected a relational expression to follow !if")

        # Scan ahead for the matching !end_if.
        ends_needed = 1
        else_idx = -1
        for idx in range(1, len(all_lines)):
            # Split the line into fields and apply text aliases.
            lno, line = all_lines[idx]
            if line == "":
                continue
            flds = self.split_line_into_fields(line)
            if flds[0] == "!if":
                ends_needed += 1
            elif flds[0] == "!else":
                if len(flds) != 1:
                    error_in_line(filename, lno, "Unexpected text after !else")
                if else_idx != -1:
                    error_in_line(filename, lno, "An !else matching line %d's !if already appeared in line %d" % (lineno, all_lines[else_idx][0]))
                else_idx = idx
            elif flds[0] == "!end_if":
                if len(flds) != 1:
                    error_in_line(filename, lno, "Unexpected text after !end_if")
                ends_needed -= 1
                if ends_needed == 0:
                    break

        # If the condition is true, parse the body.  Otherwise do nothing.
        if ends_needed != 0:
            error_in_line(filename, lineno, "Failed to find a matching !end_if directive")
        end_idx = idx
        ast = self.rel_parser.parse(filename, lineno, " ".join(self.env.sub_syms(fields[1:])))
        ast.compile()
        rhs = ast.evaluate(dict(self.env))
        if rhs:
            # Evaluate the then clause in a new scope.
            self.env.push()
            if else_idx == -1:
                # No else clause
                self.process_file_contents(filename, all_lines[1:end_idx])
            else:
                # else clause
                self.process_file_contents(filename, all_lines[1:else_idx])
            self.env.pop()
        elif else_idx != -1:
            # Evaluate the else clause in a new scope.
            self.env.push()
            self.process_file_contents(filename, all_lines[else_idx+1:end_idx])
            self.env.pop()

        # Process the rest of the file.
        self.process_file_contents(filename, all_lines[end_idx+1:])

    def process_for(self, filename, lineno, fields, all_lines):
        """Parse and process a !for directive.  Recursively parse the remaining
        file contents."""
        # Parse the !for line.
        if len(fields) < 4 or fields[2] != ":=":
            error_in_line(filename, lineno, 'Expected a variable name, an ":=", and a comma-separated list to follow !for')
        seq = [s.strip() for s in " ".join(self.env.sub_syms(fields[3:])).split(",")]
        for i in range(len(seq)):
            if seq[i] == "..." and i != len(seq) - 2:
                error_in_line(filename, lineno, '"..." can appear only in the penultimate position in a sequence')

        # Construct an iterator based on the given sequence.
        if "..." in seq:
            seq_ints = []
            for i in range(len(seq)):
                if seq[i] == "...":
                    seq_ints.append("...")
                else:
                    ast = self.expr_parser.parse(filename, lineno, seq[i])
                    ast.compile()
                    val = ast.evaluate(dict(self.env))
                    seq_ints.append(val)
            iter = LoopIterator(filename, lineno, seq_ints)
        else:
            iter = seq

        # Scan ahead for the matching !end_for.
        ends_needed = 1
        for idx in range(1, len(all_lines)):
            # Split the line into fields and apply text aliases.
            lno, line = all_lines[idx]
            if line == "":
                continue
            flds = self.split_line_into_fields(line)
            if flds[0] == "!for":
                ends_needed += 1
            elif flds[0] == "!end_for":
                if len(flds) != 1:
                    error_in_line(filename, lno, "Unexpected text after !end_for")
                ends_needed -= 1
                if ends_needed == 0:
                    break

        # Parse the body once per iterator element.
        if ends_needed != 0:
            error_in_line(filename, lineno, "Failed to find a matching !end_for directive")
        end_idx = idx
        for val in iter:
            self.env.push()
            self.env[fields[1]] = val
            self.process_file_contents(filename, all_lines[1:end_idx])
            self.env.pop()

        # Process the rest of the file.
        self.process_file_contents(filename, all_lines[end_idx+1:])


    def process_pin(self, filename, lineno, pin_str):
        "Parse a pin statement into one or more Pin objects and add these to the program."
        lhs_rhs = pin_str.split(":=")
        if len(lhs_rhs) != 2:
            self.qmasm.abend('Failed to parse pin statement "%s"' % pin_str)
        pin_parser = PinParser()
        lhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[0]))
        rhs_list = pin_parser.parse_rhs(self.env.sub_syms(lhs_rhs[1]))
        if len(lhs_list) != len(rhs_list):
            self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (pin_str, len(lhs_list), len(rhs_list)))
        return [Pin(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]

    def process_chain(self, filename, lineno, chain_str):
        "Parse a chain statement into one or more Chain objects and add these to the program."
        # We use the LHS parser from PinParser to parse both sides of the chain.
        lhs_rhs = chain_str.split("=")
        if len(lhs_rhs) != 2:
            self.qmasm.abend('Failed to parse chain statement "%s"' % chain_str)
        pin_parser = PinParser()
        lhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[0]))
        rhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[1]))  # Note use of parse_lhs to parse the RHS.
        if len(lhs_list) != len(rhs_list):
            self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (chain_str, len(lhs_list), len(rhs_list)))
        return [Chain(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]

    def process_antichain(self, filename, lineno, antichain_str):
        "Parse an anti-chain statement into one or more AntiChain objects and add these to the program."
        # We use the LHS parser from PinParser to parse both sides of the anti-chain.
        lhs_rhs = antichain_str.split("/=")
        if len(lhs_rhs) != 2:
            self.qmasm.abend('Failed to parse anti-chain statement "%s"' % antichain_str)
        pin_parser = PinParser()
        lhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[0]))
        rhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[1]))  # Note use of parse_lhs to parse the RHS.
        if len(lhs_list) != len(rhs_list):
            self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (antichain_str, len(lhs_list), len(rhs_list)))
        return [AntiChain(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]

    def process_alias(self, filename, lineno, alias_str):
        "Parse an alias statement into one or more Alias objects and add these to the program."
        # We use the LHS parser from PinParser to parse both sides of the alias.
        lhs_rhs = alias_str.split("<->")
        if len(lhs_rhs) != 2:
            self.qmasm.abend('Failed to parse alias statement "%s"' % alias_str)
        pin_parser = PinParser()
        lhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[0]))
        rhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[1]))  # Note use of parse_lhs to parse the RHS.
        if len(lhs_list) != len(rhs_list):
            self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (alias_str, len(lhs_list), len(rhs_list)))
        return [Alias(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]

    def process_file_contents(self, filename, all_lines):
        """Parse the contents of a file.  Contents are passed as a list plus an
        initial line number."""
        for idx in range(len(all_lines)):
            # Split the line into fields and apply text aliases.
            lineno, line = all_lines[idx]
            if line == "":
                continue
            fields = self.split_line_into_fields(line)
            nfields = len(fields)
            if nfields == 0:
                continue

            # Process the line.
            if fields[0] == "!if":
                # Special case for !if directives
                return self.process_if(filename, lineno, fields, all_lines[idx:])
            elif fields[0] == "!for":
                # Special case for !for directives
                return self.process_for(filename, lineno, fields, all_lines[idx:])
            try:
                # Parse first-field directives.
                func = self.dir_to_func[fields[0]]
            except KeyError:
                # Reject unrecognized directives.
                if fields[0][0] == "!":
                    if fields[0] in ["!else", "!end_if"]:
                        error_in_line(filename, lineno, "Encountered an %s without a matching !if" % fields[0])
                    elif fields[0] == "!end_for":
                        error_in_line(filename, lineno, "Encountered an !end_for without a matching !for")
                    else:
                        error_in_line(filename, lineno, "Unrecognized directive %s" % fields[0])

                # Prohibit "!next." outside of macros.
                if self.current_macro[0] == None:
                    for f in fields:
                        if "!next." in f:
                            error_in_line(filename, lineno, '"!next." is allowed only within !begin_macro...!end_macro blocks')

                # Parse all lines not containing a directive in the first field.
                if nfields == 2:
                    func = self.parse_line_weight
                elif nfields == 3 and fields[1] == "=":
                    func = self.parse_line_chain
                elif nfields == 3 and fields[1] == "/=":
                    func = self.parse_line_antichain
                elif nfields == 3 and fields[1] == ":=":
                    func = self.parse_line_pin
                elif nfields == 3 and fields[1] == "<->":
                    func = self.parse_line_alias
                elif nfields >= 3 and "->" in fields:
                    func = self.parse_line_rename
                elif nfields == 3 and self.is_float(fields[2]):
                    func = self.parse_line_strength
                else:
                    # None of the above
                    error_in_line(filename, lineno, 'Failed to parse "%s"' % line)
            func(filename, lineno, fields)

    def process_file(self, filename, infile):
        """Define a function that parses an input file into an internal
        representation.  This function can be called recursively (due to !include
        directives)."""

        # Read the entire file into a list.
        all_lines = []
        lineno = 1
        for line in infile:
            all_lines.append((lineno, line.strip()))
            lineno += 1
        self.process_file_contents(filename, all_lines)

    def process_files(self, file_list):
        "Parse a list of file(s) into an internal representation."
        if file_list == []:
            # No files were specified: Read from standard input.
            self.process_file("<stdin>", sys.stdin)
            if self.current_macro[0] != None:
                error_in_line(filename, lineno, "Unterminated definition of macro %s" % self.current_macro[0])
        else:
            # Files were specified: Process each in turn.
            for infilename in file_list:
                try:
                    infile = open(infilename)
                except IOError:
                    self.qmasm.abend('Failed to open %s for input' % infilename)
                self.process_file(infilename, infile)
                if self.current_macro[0] != None:
                    error_in_line(filename, lineno, "Unterminated definition of macro %s" % self.current_macro[0])
                infile.close()

class PinParser(object):
    "Provide methods for parsing a pin statement."

    def __init__(self):
        self.bracket_re = re.compile(r'^\s*(\d+)(\s*(?:\.\.|:)\s*(\d+))?\s*$')
        self.bool_re = re.compile(r'TRUE|FALSE|T|F|0|[-+]?1', re.IGNORECASE)

        # Define synonyms for "true" and "false".
        self.str2bool = {s: True for s in ["1", "+1", "T", "TRUE"]}
        self.str2bool.update({s: False for s in ["0", "-1", "F", "FALSE"]})

    def expand_brackets(self, vars, expr):
        """Repeat one or more variables for each bracketed expression.  For
        example, expanding ("hello", "1 .. 3") should produce
        ("hello[1]", "hello[2]", "hello[3]")."""
        # Determine the starting and ending numbers and the step.
        bmatch = self.bracket_re.search(expr)
        if bmatch == None:
            return ["%s[%s]" % (v, expr) for v in vars]
        bmatches = bmatch.groups()
        num1 = int(bmatches[0])
        if bmatches[2] == None:
            num2 = num1
        else:
            num2 = int(bmatches[2])
        if num1 <= num2:
            step = 1
        else:
            step = -1

        # Append the same bracketed constant to each variable.
        new_vars = []
        for v in vars:
            for i in range(num1, num2 + step, step):
                new_vars.append("%s[%d]" % (v, i))
        return new_vars

    def parse_lhs(self, lhs):
        "Parse the left-hand side of a pin statement."
        variables = [""]
        group_len = 1    # Number of variables produced from the same bracketed expression
        bracket_expr = ""
        in_bracket = False
        for c in lhs:
            if c == "[":
                if in_bracket:
                    self.qmasm.abend("Nested brackets are not allowed")
                in_bracket = True
            elif c == "]":
                if not in_bracket:
                    self.qmasm.abend('Encountered "]" before seeing a "["')
                old_vars = variables[:-group_len]
                current_vars = variables[-group_len:]
                new_vars = self.expand_brackets(current_vars, bracket_expr)
                variables = old_vars + new_vars
                group_len = len(new_vars)
                in_bracket = False
                bracket_expr = ""
            elif in_bracket:
                bracket_expr += c
            elif c == " " or c == "\t":
                if in_bracket:
                    self.qmasm.abend("Unterminated bracketed expression")
                if variables[-1] != "":
                    variables.append("")
                group_len = 1
            else:
                for i in range(1, group_len + 1):
                    variables[-i] += c
        if in_bracket:
            self.qmasm.abend("Unterminated bracketed expression")
        if variables[-1] == "":
            variables.pop()
        return variables

    def parse_rhs(self, rhs):
        "Parse the right-hand side of a pin statement."
        for inter in [t.strip() for t in self.bool_re.split(rhs)]:
            if inter != "":
                self.qmasm.abend('Unexpected "%s" in pin right-hand side "%s"' % (inter, rhs))
        return [self.str2bool[t.upper()] for t in self.bool_re.findall(rhs)]



###################################
# Solve a binary quadratic model  #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import copy
import dimod
import hashlib
import json
import marshal
import minorminer
import os
import re
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from dimod import ExactSolver, SampleSet
from dwave.cloud import Client
from dwave.embedding import embed_bqm
from dwave.system import DWaveSampler, EmbeddingComposite
from dwave_qbsolv import QBSolv
from neal import SimulatedAnnealingSampler
from qmasm.solutions import Solutions
from tabu import TabuSampler

class EmbeddingCache(object):
    "Read and write an embedding cache file."

    def __init__(self, qmasm, edges, adj):
        # Ensure we have a valid cache directory.
        self.hash = None
        try:
            self.cachedir = os.environ["QMASMCACHE"]
        except KeyError:
            self.cachedir = None
            return None
        if not os.path.isdir(self.cachedir):
            qmasm.abend("QMASMCACHE is set to %s, which is not an extant directory" % self.cachedir)

        # Compute a SHA-1 sum of our inputs.
        sha = hashlib.sha1()
        sha.update(str(sorted(edges)).encode("utf-8"))
        sha.update(str(sorted(adj)).encode("utf-8"))
        self.hash = sha.hexdigest()

    def read(self):
        "Read an embedding from an embedding cache or None on a cache miss."
        if self.hash == None:
            return None
        try:
            h = open(os.path.join(self.cachedir, self.hash), "rb")
        except IOError:
            return None
        embedding = marshal.load(h)
        h.close()
        return embedding

    def write(self, embedding):
        "Write an embedding to an embedding cache."
        if self.hash == None:
            return
        try:
            h = open(os.path.join(self.cachedir, self.hash), "wb")
        except IOError:
            return None
        marshal.dump(embedding, h)
        h.close()

class Sampler(object):
    "Interface to Ocean samplers."

    # Keep track of parameters rejected by the solver.
    unexp_arg_re = re.compile(r"got an unexpected keyword argument '(\w+)'")
    rejected_params = []

    def __init__(self, qmasm, profile=None, solver=None, qbsolv=False):
        "Acquire either a software sampler or a sampler representing a hardware solver."
        self.qmasm = qmasm
        self.sampler, self.client_info = self._get_sampler(profile, solver)
        if qbsolv:
            if solver == None:
                self.qbsolv_sampler = EmbeddingComposite(self.sampler)
            else:
                self.qbsolv_sampler = self.sampler
            self.sampler = QBSolv()
            self.client_info["solver_name"] = "QBSolv + " + self.client_info["solver_name"]
        else:
            self.qbsolv_sampler = None

    def _get_sampler(self, profile, solver):
        "Return a dimod.Sampler object."
        # Handle built-in software samplers as special cases.
        info = {}
        if solver != None:
            info["solver_name"] = solver
        if solver == "exact":
            return ExactSolver(), info
        elif solver == "neal":
            return SimulatedAnnealingSampler(), info
        elif solver == "tabu":
            return TabuSampler(), info

        # In the common case, read the configuration file, either the
        # default or the one named by the DWAVE_CONFIG_FILE environment
        # variable.
        if profile != None:
            info["profile"] = profile
        try:
            with Client.from_config(profile=profile) as client:
                if solver == None:
                    solver = client.default_solver
                else:
                    solver = {"name": solver}
                sampler = DWaveSampler(profile=profile, solver=solver)
                info = {"solver_name": sampler.solver.name,
                        "endpoint": client.endpoint}
                return sampler, info
        except Exception as err:
            self.qmasm.abend("Failed to construct a sampler (%s)" % str(err))

    def show_properties(self, verbose):
        "Output either short or all solver properties."
        if verbose == 0:
            return

        # Combine properties from various sources into a single dictionary.
        props = self.client_info.copy()
        if self.qbsolv_sampler != None:
            props.update(self.qbsolv_sampler.properties)
        else:
            props.update(self.sampler.properties)

        # Determine the width of the widest key.
        max_key_len = len("Parameter")
        prop_keys = sorted(props)
        for k in prop_keys:
            max_key_len = max(max_key_len, len(k))

        # Output either "short" values (if verbose = 1) or all values (if
        # verbose > 1).
        short_value_len = 70 - max_key_len
        sys.stderr.write("Encountered the following solver properties:\n\n")
        sys.stderr.write("    %-*s  Value\n" % (max_key_len, "Parameter"))
        sys.stderr.write("    %s  %s\n" % ("-" * max_key_len, "-" * max(5, short_value_len)))
        for k in prop_keys:
            if isinstance(props[k], str):
                val_str = props[k]
            else:
                val_str = repr(props[k])
            if verbose >= 2 or len(val_str) <= short_value_len:
                sys.stderr.write("    %-*s  %s\n" % (max_key_len, k, val_str))
            elif len(val_str) > short_value_len:
                # Value is too long for a single -v.
                if isinstance(props[k], list):
                    val_str = "(%d items; use -v -v to view)" % len(props[k])
                else:
                    val_str = "(%d characters; use -v -v to view)" % len(val_str)
                sys.stderr.write("    %-*s  %s\n" % (max_key_len, k, val_str))
        sys.stderr.write("\n")

    def _find_adjacency(self, sampler):
        "Perform a depth-first search for an adjacency attribute."
        # Successful base case: The given sampler reports its adjacency
        # structure.
        try:
            return sampler.adjacency
        except AttributeError:
            pass

        # Failed base case: The given sampler has no children.
        try:
            children = sampler.children
        except AttributeError:
            return None

        # Recursive case: Search each child for its adjacency.
        for c in children:
            adj = self._find_adjacency(c)
            if adj != None:
                return adj
        return None

    def get_hardware_adjacency(self):
        "Return the hardware adjacency structure, if any."
        return self._find_adjacency(self.sampler)

    def read_hardware_adjacency(self, fname, verbosity):
        """Read a hardware adjacency list from a file.  Each line must contain
        a space-separated pair of vertex numbers."""
        # Read from a file to a set of vertex pairs.
        adj = set()
        lineno = 0
        if verbosity >= 2:
            sys.stderr.write("Reading hardware adjacency from %s.\n" % fname)
        with open(fname) as f:
            for orig_line in f:
                # Discard comments then parse the line into exactly two vertices.
                lineno += 1
                line = orig_line.partition("#")[0]
                try:
                    verts = [int(v) for v in line.split()]
                except ValueError:
                    self.qmasm.abend('Failed to parse line %d of file %s ("%s")' % (lineno, fname, orig_line.strip()))
                if len(verts) == 0:
                    continue
                if len(verts) != 2 or verts[0] == verts[1]:
                    self.qmasm.abend('Failed to parse line %d of file %s ("%s")' % (lineno, fname, orig_line.strip()))

                # Canonicalize the vertex numbers and add the result to the set.
                if verts[1] > verts[0]:
                    verts[0], verts[1] = verts[1], verts[0]
                adj.add((verts[0], verts[1]))
        if verbosity >= 2:
            sys.stderr.write("%d unique edges found\n\n" % len(adj))

        # Convert from vertex pairs to a dictionary from a vertex to its
        # neighbors.  Note that we treat all edges as bidirectional.
        adj_dict = {}
        for u, v in adj:
            try:
                adj_dict[u].append(v)
            except KeyError:
                adj_dict[u] = [v]
            try:
                adj_dict[v].append(u)
            except KeyError:
                adj_dict[v] = [u]
        return adj_dict

    def _find_embedding(self, edges, adj, **kwargs):
        "Wrap minorminer.find_embedding with a version that intercepts its output."
        # Minorminer accepts the hardware adjacency as a list of
        # pairs, not a map from each node to its neighbors.
        mm_adj = [(u, v) for u in adj for v in adj[u]]

        # Given verbose=0, invoke minorminer directly.
        if "verbose" not in kwargs or kwargs["verbose"] == 0:
            # Convert all keys to integers for consistency.
            embedding = minorminer.find_embedding(edges, mm_adj, **kwargs)
            return {int(k): v for k, v in embedding.items()}

        # minorminer's find_embedding is hard-wired to write to stdout.
        # Trick it into writing into a pipe instead.
        sepLine = "=== EMBEDDING ===\n"
        r, w = os.pipe()
        pid = os.fork()
        if pid == 0:
            # Child -- perform the embedding.
            os.close(r)
            os.dup2(w, sys.stdout.fileno())
            embedding = minorminer.find_embedding(edges, mm_adj, **kwargs)
            sys.stdout.flush()
            os.write(w, sepLine.encode())
            os.write(w, (json.dumps(embedding) + "\n").encode())
            os.close(w)
            os._exit(0)
        else:
            # Parent -- report the embedding's progress.
            os.close(w)
            pipe = os.fdopen(r, "r", 10000)
            while True:
                try:
                    rstr = pipe.readline()
                    if rstr == sepLine:
                        break
                    if rstr == "":
                        self.qmasm.abend("Embedder failed to terminate properly")
                    sys.stderr.write("      %s" % rstr)
                except:
                    pass

            # Receive the embedding from the child.  Convert all keys to
            # integers for consistency.
            embedding = json.loads(pipe.readline())
            return {int(k): v for k, v in embedding.items()}

    def find_problem_embedding(self, logical, topology_file, verbosity):
        """Find an embedding of a problem on a physical topology, if
        necessary.  Return a physical Sampler object."""
        # Create a physical Problem.
        physical = copy.deepcopy(logical)
        physical.embedding = {}

        # Acquire the hardware topology unless we were given a specific
        # topology to use.
        if topology_file == None:
            hw_adj = self.get_hardware_adjacency()
        else:
            hw_adj = self.read_hardware_adjacency(topology_file, verbosity)
        physical.hw_adj = hw_adj

        # See if we already have an embedding in the embedding cache.
        edges = logical.bqm.quadratic
        if hw_adj == None or len(edges) == 0:
            # Either the sampler does not require embedding, or we have no work
            # to do.
            return physical
        if verbosity >= 2:
            sys.stderr.write("Minor-embedding the logical problem onto the physical topology:\n\n")
        ec = EmbeddingCache(self.qmasm, edges, hw_adj)
        if verbosity >= 2:
            if ec.cachedir == None:
                sys.stderr.write("  No embedding cache directory was specified ($QMASMCACHE).\n")
            else:
                sys.stderr.write("  Using %s as the embedding cache directory.\n" % ec.cachedir)
        embedding = ec.read()
        if embedding == {}:
            # Cache hit, but embedding had failed
            if verbosity >= 2:
                sys.stderr.write("  Found failed embedding %s in the embedding cache.\n\n" % ec.hash)
        elif embedding != None:
            # Successful cache hit!
            if verbosity >= 2:
                sys.stderr.write("  Found successful embedding %s in the embedding cache.\n\n" % ec.hash)
            physical.embedding = embedding
            return physical
        if verbosity >= 2 and ec.cachedir != None:
            sys.stderr.write("  No existing embedding found in the embedding cache.\n")

        # Minor-embed the logical problem onto the hardware topology.
        if verbosity < 2:
            physical.embedding = self._find_embedding(edges, hw_adj)
        else:
            sys.stderr.write("  Running the embedder.\n\n")
            physical.embedding = self._find_embedding(edges, hw_adj, verbose=1)
            sys.stderr.write("\n")

        # Cache the embedding for next time.
        ec.write(physical.embedding)
        if verbosity >= 2 and ec.cachedir != None:
            sys.stderr.write("  Caching the embedding as %s.\n\n" % ec.hash)
        return physical

    def embed_problem(self, logical, topology_file, verbosity):
        "Embed a problem on a physical topology, if necessary."
        # Embed the problem.  We first filter out isolated variables that don't
        # appear in the embedding graph to prevent embed_bqm from complaining.
        physical = self.find_problem_embedding(logical, topology_file, verbosity)
        if physical.embedding == {}:
            # No embedding is necessary.
            physical.logical = logical
            return physical
        physical.bqm.remove_variables_from([q
                                            for q in physical.bqm.linear
                                            if q not in physical.embedding and physical.bqm.linear[q] == 0.0])
        for q, wt in physical.bqm.linear.items():
            if q not in physical.embedding and wt != 0.0:
                self.qmasm.abend("Logical qubit %d has a nonzero weight (%.5g) but was not embedded" % (q, wt))
        physical.bqm = embed_bqm(physical.bqm, physical.embedding,
                                 physical.hw_adj, -physical.qmasm.chain_strength)

        # Update weights and strengths.  Maintain a reference to the
        # logical problem.
        physical.logical = logical
        physical.weights = physical.bqm.linear
        physical.strengths = physical.bqm.quadratic

        # Some problem parameters are not relevant in the physical problem.
        physical.chains = None
        physical.antichains = None
        physical.pinned = None
        physical.known_values = None
        return physical

    def _get_default_annealing_time(self):
        "Determine a suitable annealing time to use if none was specified."
        try:
            # Use the default value.
            anneal_time = self.sampler.properties["default_annealing_time"]
        except KeyError:
            try:
                # If the default value is undefined, use the minimum allowed
                # value.
                anneal_time = self.sampler.properties["annealing_time_range"][0]
            except KeyError:
                # If all else fails, use 20 as a reasonable default.
                anneal_time = 20
        return anneal_time

    def _compute_sample_counts(self, samples, anneal_time):
        "Return a list of sample counts to request of the hardware."
        # The formula in the D-Wave System Documentation (under
        # "problem_run_duration_range") is Duration = (annealing_time +
        # readout_thermalization)*num_reads + programming_thermalization.  We
        # split the number of samples so that each run time is less than the
        # maximum duration.
        props = self.sampler.properties
        try:
            max_run_duration = props["problem_run_duration_range"][1]
            prog_therm = props["default_programming_thermalization"]
            read_therm = props["default_readout_thermalization"]
        except KeyError:
            # Assume we're on a software solver.
            return [samples]
        max_samples = (max_run_duration - prog_therm)//(anneal_time + read_therm)
        # Independent of the maximum sample count just computed, the number of
        # samples can't exceed the maximum the hardware allows.
        try:
            max_samples = min(max_samples, props["num_reads_range"][1])
        except KeyError:
            pass

        # Split the number of samples into pieces of size max_samples.
        if samples <= max_samples:
            samples_list = [samples]
        else:
            samples_list = []
            s = samples
            while s > 0:
                if s >= max_samples:
                    samples_list.append(max_samples)
                    s -= max_samples
                else:
                    samples_list.append(s)
                    s = 0
        return samples_list

    def _compute_spin_rev_counts(self, spin_revs, samples_list):
        "Divide a total number of spin reversals across a set of samples."
        samples = sum(samples_list)
        spin_rev_frac = float(spin_revs)/float(samples)  # We'll evenly divide our spin reversals among samples.
        spin_rev_list = [int(samps*spin_rev_frac) for samps in samples_list]
        missing_srs = spin_revs - sum(spin_rev_list)
        while missing_srs > 0:
            # Account for rounding errors by adding back in some missing spin
            # reversals.
            for i in range(samples):
                if samples_list[i] > spin_rev_list[i]:
                    spin_rev_list[i] += 1
                    missing_srs -= 1
                    if missing_srs == 0:
                        break
        return spin_rev_list

    def _merge_results(self, results):
        "Merge results into a single SampleSet."
        sum_keys = ["total_real_time",
                    "qpu_access_overhead_time",
                    "post_processing_overhead_time",
                    "qpu_sampling_time",
                    "total_post_processing_time",
                    "qpu_programming_time",
                    "run_time_chip",
                    "qpu_access_time"]
        avg_keys = ["anneal_time_per_run",
                    "readout_time_per_run",
                    "qpu_delay_time_per_sample",
                    "qpu_anneal_time_per_sample",
                    "qpu_readout_time_per_sample"]
        timing = {}
        merged = dimod.concatenate(results)
        nsamples = len(merged)
        for sk in sum_keys:
            try:
                timing[sk] = sum([r.info["timing"][sk] for r in results])
            except KeyError:
                pass
        for ak in avg_keys:
            try:
                timing[ak] = sum([r.info["timing"][ak]*len(r) for r in results])//nsamples
            except KeyError:
                pass
        merged.info["timing"] = timing
        return merged

    def _submit_and_block(self, bqm, verbosity, **params):
        "Submit a job and wait for it to complete"
        # This method is a workaround for a bug in dwave-system.  See
        # https://github.com/dwavesystems/dwave-system/issues/297#issuecomment-632384524
        sampler = self.sampler
        sub_params = copy.copy(params)
        if self.qbsolv_sampler != None:
            sub_params["solver"] = self.qbsolv_sampler
            if verbosity >= 2:
                sub_params["verbosity"] = 10   # Arbitrary large number
        result = None
        while result == None:
            # I don't know of a way to query a solver for the parameters it
            # accepts.  Hence, we resort to the grotesque hack of parsing the
            # error string to determine what is and isn't acceptable.
            try:
                result = sampler.sample(bqm, **sub_params)
            except TypeError as err:
                match = self.unexp_arg_re.search(str(err))
                if match == None:
                    raise
                p = match[1]
                self.rejected_params.append(p)
                del sub_params[p]
        result.resolve()
        return result

    def acquire_samples(self, verbosity, physical, samples, anneal_time, spin_revs, postproc):
        "Acquire a number of samples from either a hardware or software sampler."
        # Map abbreviated to full names for postprocessing types.
        postproc = {"none": "", "opt": "optimization", "sample": "sampling"}[postproc]

        # Determine the annealing time to use.
        if anneal_time == None:
            anneal_time = self._get_default_annealing_time()

        # Compute a list of the number of samples to take each iteration
        # and the number of spin reversals to perform.
        samples_list = self._compute_sample_counts(samples, anneal_time)
        spin_rev_list = self._compute_spin_rev_counts(spin_revs, samples_list)
        nqmis = len(samples_list)   # Number of (non-unique) QMIs to submit

        # QBSolv writes to standard output, but we want it to write to standard
        # error instead.  Note that the following is imperfect because of C
        # buffering.  Setting PYTHONUNBUFFERED=1 in the environment seems to
        # help, though.
        if self.qbsolv_sampler != None:
            stdout_fileno = os.dup(sys.stdout.fileno())
            os.dup2(sys.stderr.fileno(), sys.stdout.fileno())

        # Submit all of our QMIs asynchronously.
        results = [None for i in range(nqmis)]
        executor = ThreadPoolExecutor()
        overall_start_time = time.time_ns()
        for i in range(nqmis):
            solver_params = dict(chains=list(physical.embedding.values()),
                                 num_reads=samples_list[i],
                                 annealing_time=anneal_time,
                                 num_spin_reversal_transforms=spin_rev_list[i],
                                 postprocess=postproc)
            for p in self.rejected_params:
                del solver_params[p]
            future = executor.submit(self._submit_and_block, physical.bqm, verbosity, **solver_params)
            results[i] = SampleSet.from_future(future)

        # Wait for the QMIs to finish.
        executor.shutdown(wait=False)
        if verbosity == 1:
            if nqmis == 1:
                sys.stderr.write("Waiting for the problem to complete.\n\n")
            else:
                sys.stderr.write("Waiting for all %d subproblems to complete.\n\n" % nqmis)
        elif verbosity >= 2:
            sys.stderr.write("Number of subproblems completed:\n\n")
            cdigits = len(str(nqmis))     # Digits in the number of completed QMIs
            tdigits = len(str(nqmis*5))   # Estimate 5 seconds per QMI submission
            start_time = time.time_ns()
        ncomplete = 0
        prev_ncomplete = 0
        while ncomplete < nqmis:
            ncomplete = sum([int(r.done()) for r in results])
            if verbosity >= 2 and ncomplete > prev_ncomplete:
                end_time = time.time_ns()
                sys.stderr.write("    %*d of %d (%3.0f%%) after %*.0f second(s)\n" %
                                 (cdigits, ncomplete, nqmis,
                                  100.0*float(ncomplete)/float(nqmis),
                                  tdigits, (end_time - start_time)/1e9))
                prev_ncomplete = ncomplete
            if ncomplete < nqmis:
                time.sleep(1)
        overall_end_time = time.time_ns()
        if verbosity >= 2:
            sys.stderr.write("\n")
            sys.stderr.write("    Average time per subproblem: %.2g second(s)\n\n" % ((overall_end_time - overall_start_time)/(nqmis*1e9)))
            if "problem_id" in results[0].info:
                sys.stderr.write("IDs of completed subproblems:\n\n")
                for i in range(nqmis):
                    sys.stderr.write("    %s\n" % results[i].info["problem_id"])
                sys.stderr.write("\n")

        # Merge the result of seperate runs into a composite answer.
        answer = self._merge_results(results)
        answer.info["timing"]["round_trip_time"] = (overall_end_time - overall_start_time)//1000

        # Reset standard output.
        if self.qbsolv_sampler != None:
            os.dup2(stdout_fileno, sys.stdout.fileno())

        # Return a Solutions object for further processing.
        return Solutions(answer, physical, verbosity >= 2)



#! /usr/bin/env python

###################################
# Quantum Macro Assembler         #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import re
import sys
from qmasm.cmdline import ParseCommandLine
from qmasm.output import OutputMixin
from qmasm.parse import FileParser
from qmasm.problem import Problem
from qmasm.solve import Sampler
from qmasm.utils import Utilities, SymbolMapping

class QMASM(ParseCommandLine, Utilities, OutputMixin):
    "QMASM represents everything the program can do."

    def __init__(self):
        # List of Statement objects
        self.program = []

        # Map between symbols and numbers.
        self.sym_map = SymbolMapping()

        # Multiple components of QMASM require a definition of an identifier.
        self.ident_re = re.compile(r'[^-+*/%&\|^~()<=>#,\s]+')

        # Define a scale factor for converting floats to ints for MiniZinc's
        # sake.
        self.minizinc_scale_factor = 10000.0

    def run(self):
        "Execute the entire QMASM processing sequence."

        # Parse the command line.
        cl_args = self.parse_command_line()
        self.report_command_line(cl_args)

        # Parse the original input file(s) into an internal representation.
        fparse = FileParser(self)
        fparse.process_files(cl_args.input)

        # Parse the variable pinnings specified on the command line.  Append
        # these to the program.
        if cl_args.pin != None:
            for pin in cl_args.pin:
                self.program.extend(fparse.process_pin("[command line]", 1, pin))

        # Walk the statements in the program, processing each in turn.
        logical = Problem(self, cl_args.qubo)
        for stmt in self.program:
            stmt.update_qmi("", "<ERROR>", logical)

        # Define a strength for each user-specified chain and anti-chain, and
        # assign strengths to those chains.
        self.chain_strength = logical.assign_chain_strength(cl_args.chain_strength)
        if cl_args.verbose >= 1:
            sys.stderr.write("Chain strength: %7.4f\n\n" % self.chain_strength)

        # We now have enough information to produce an Ocean BinaryQuadraticModel.
        logical.generate_bqm()

        # Convert chains to aliases where possible.
        if cl_args.O >= 1:
            logical.convert_chains_to_aliases(cl_args.verbose)

        # Simplify the problem if possible.
        if cl_args.O >= 1:
            logical.simplify_problem(cl_args.verbose)

        # Establish a connection to a D-Wave or software sampler.
        sampler = Sampler(self, profile=cl_args.profile, solver=cl_args.solver, qbsolv=cl_args.qbsolv)
        sampler.show_properties(cl_args.verbose)

        # Convert user-specified chains, anti-chains, and pins to assertions.
        logical.append_assertions_from_statements()

        # Determine if we're expected to write an output file.  If --run was
        # specified, we write a file only if --output was also specified.
        write_output_file = not (cl_args.output == "<stdout>" and cl_args.run)

        # Determine when to write an output file: either pre- or post-embedding.
        write_time = None
        if write_output_file:
            # Start with always_embed --> post, otherwise pre.
            write_time = "pre"
            if cl_args.always_embed:
                write_time = "post"

            # QMASM output is always pre-embedding.
            if cl_args.format == "qmasm" and write_time == "post":
                self.warn("Ignoring --always embed; incompatible with --format=qmasm")
                write_time = "pre"

            # Qubist output is always post-embedding.
            if cl_args.format == "qubist":
                write_time = "post"

        # Produce pre-embedding output files.
        if write_output_file and write_time == "pre":
            self.write_output(logical, cl_args.output, cl_args.format, cl_args.qubo, sampler)
            if not cl_args.run:
                sys.exit(0)

        # Embed the problem on the physical topology.
        physical = sampler.embed_problem(logical, cl_args.topology_file, cl_args.verbose)

        # Map each logical qubit to one or more symbols.
        max_num = self.sym_map.max_number()
        num2syms = [[] for _ in range(max_num + 1)]
        all_num2syms = [[] for _ in range(max_num + 1)]
        max_sym_name_len = 7
        for s, n in self.sym_map.symbol_number_items():
            all_num2syms[n].append(s)
            if cl_args.verbose >= 2 or "$" not in s:
                num2syms[n].append(s)
                max_sym_name_len = max(max_sym_name_len, len(repr(num2syms[n])) - 1)

        # Output the embedding.
        physical.output_embedding(cl_args.verbose, max_sym_name_len, num2syms)

        # Abort if any variables failed to embed.
        if physical.embedding != {}:
            danglies = physical.dangling_variables(num2syms)
            if len(danglies) > 0:
                self.abend("Disconnected variables encountered: %s" % str(sorted(danglies)))

        # Output some problem statistics.
        if cl_args.verbose > 0:
            physical.output_embedding_statistics()

        # Produce post-embedding output files.
        if write_output_file and write_time == "post":
            self.write_output(physical, cl_args.output, cl_args.format, cl_args.qubo, sampler)

        # If we weren't told to run anything we can exit now.
        if not cl_args.run:
            return

        # Solve the specified problem.
        if cl_args.verbose >= 1:
            sys.stderr.write("Submitting the problem to the %s solver.\n\n" % sampler.client_info["solver_name"])
        solutions = sampler.acquire_samples(cl_args.verbose,
                                            physical,
                                            cl_args.samples,
                                            cl_args.anneal_time,
                                            cl_args.spin_revs,
                                            cl_args.postproc)
        solutions.report_timing_information(cl_args.verbose)
        solutions.report_chain_break_information(cl_args.verbose)

        # Filter the solutions as directed by the user.
        filtered_solns = solutions.filter(cl_args.show, cl_args.verbose, cl_args.samples)

        # Output energy tallies.
        if cl_args.verbose >= 2:
            solutions.report_energy_tallies(filtered_solns, True)
        elif cl_args.verbose >= 1:
            solutions.report_energy_tallies(filtered_solns, False)

        # Output the solution to the standard output device.
        show_asserts = cl_args.verbose >= 2 or cl_args.show in ["best", "all"]
        filtered_solns.output_solutions(cl_args.values, cl_args.verbose, show_asserts)

def main():
    "Run QMASM."
    q = QMASM()
    q.run()

if __name__ == "__main__":
    main()



###################################
# Define an Ising or QUBO problem #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import copy
import dimod
import sys
from collections import defaultdict
from qmasm.assertions import AssertParser

# Problem is currently just a thin veneer over dimod.BinaryQuadraticModel.  If
# it turns out we don't even need this veneer, we may replace it with direct
# use of dimod.BinaryQuadraticModel.
class Problem(object):
    "Represent either an Ising or QUBO problem."

    def __init__(self, qmasm, qubo):
        self.qmasm = qmasm   # Pointer to the top-level QMASM class
        self.qubo = qubo     # True=QUBO; False=Ising
        self.weights = defaultdict(lambda: 0.0)    # Map from a spin to a point weight
        self.strengths = defaultdict(lambda: 0.0)  # Map from a pair of spins to a coupler strength
        self.chains = set()       # Subset of strengths keys that represents user-defined chains (always logical)
        self.antichains = set()   # Subset of strengths keys that represents user-defined anti-chains (always logical)
        self.assertions = []      # List of assertions (as ASTs) to enforce
        self.pending_asserts = [] # List of {string, op, string} tuples pending conversion to assertions
        self.pinned = []          # Pairs of {unique number, Boolean} to pin
        self.known_values = {}    # Map from a spin to -1 or +1
        self.bqm_vars = None      # Set of all variables appearing in the BQM
        self.contractions = {}    # Map from a spin to another spin it must be equal to

    def assign_chain_strength(self, ch_str):
        """Define a strength for each user-specified and automatically
        generated chain, and assign strengths to those chains (and negative
        strength to all anti-chains).  Return the computed chain strength."""
        chain_strength = ch_str
        if chain_strength == None:
            # Chain strength defaults to twice the maximum strength in the data.
            try:
                chain_strength = -2*max([abs(w) for w in self.strengths.values()])
            except ValueError:
                # No strengths -- use weights instead.
                try:
                    chain_strength = -2*max([abs(w) for w in self.weights.values()])
                except ValueError:
                    # No weights or strengths -- arbitrarily choose -1.
                    chain_strength = -1.0
        elif self.qubo:
            # With QUBO input we need to divide the chain strength by 4 for
            # consistency with the other coupler strengths.
            chain_strength /= 4.0
        for c in self.chains:
            self.strengths[c] += chain_strength
        for c in self.antichains:
            self.strengths[c] -= chain_strength
        return chain_strength

    def generate_bqm(self):
        "Generate a BinaryQuadraticModel version of the Problem."
        # Create a BQM.
        btype = dimod.SPIN
        if self.qubo:
            btype = dimod.BINARY
        bqm = dimod.BinaryQuadraticModel(self.weights, self.strengths, 0, btype, problem=self)
        if self.qubo:
            bqm.change_vartype(dimod.SPIN, inplace=True)

        # Pin all variables the user asked to pin.
        bool2spin = {False: -1, True: +1}
        pins = {q: bool2spin[b] for q, b in self.pinned}
        for q in pins:
            # Ensure that every pinned variable exists.  Otherwise,
            # fix_variables will throw a KeyError.
            bqm.add_variable(q, 0)
        bqm.fix_variables(pins)

        # Store the BQM.
        self.bqm = bqm

    def convert_to_qubo(self):
        "Return a copy of the problem with QUBO weights and strengths."
        qprob = copy.deepcopy(self)
        qprob.bqm.change_vartype(dimod.BINARY)
        qprob.weights = qprob.bqm.linear
        qprob.strengths = qprob.bqm.quadratic
        return qprob

    def convert_to_ising(self):
        "Return a copy of the problem with Ising weights and strengths."
        iprob = copy.deepcopy(self)
        iprob.bqm.change_vartype(dimod.SPIN)
        iprob.weights = iprob.bqm.linear
        iprob.strengths = iprob.bqm.quadratic
        return iprob

    def all_bqm_variables(self, force_recompute=False):
        "Return a set of all variables, referenced in linear and/or quadratic terms in the BQM."
        if self.bqm_vars != None and not force_recompute:
            return self.bqm_vars
        self.bqm_vars = set(self.bqm.linear)
        for u, v in self.bqm.quadratic:
            self.bqm_vars.add(u)
            self.bqm_vars.add(v)
        return self.bqm_vars

    def _edges_to_adj_list(self, edges):
        "Convert a list of edges to an adjacency list."
        adj = defaultdict(lambda: [])
        for u, v in edges:
            adj[u].append(v)
            adj[v].append(u)
        return adj

    def _traversal_from_root(self, adj, visited, root):
        """"Return a reversed traversal order of an adjacency list from a
        given root such that each right vertex is a leaf if all
        preceding right vertices are removed."""
        order = []
        new_visited = visited.copy()
        new_visited.add(root)
        for v in adj[root]:
            if v in new_visited:
                continue
            order.append((root, v))
            ord, vis = self._traversal_from_root(adj, new_visited, v)
            order.extend(ord)
            new_visited.update(vis)
        return order, new_visited

    def traversal_order(self, edges):
        """"Return a reversed traversal order of a graph such that each right
        vertex is a leaf if all preceding right vertices are removed."""
        adj = self._edges_to_adj_list(edges)
        order = []
        nodes = set()
        for u, v in edges:
            nodes.add(u)
            nodes.add(v)
        visited = set()
        for u in nodes:
            if u in visited:
                continue
            ord, vis = self._traversal_from_root(adj, visited, u)
            order.extend(ord)
            visited.update(vis)
        order.reverse()
        return order

    def convert_chains_to_aliases(self, verbosity):
        "Replace user-defined chains with aliases."
        # Say what we're about to do
        if verbosity >= 2:
            sys.stderr.write("Replaced user-defined chains with aliases:\n\n")
            sys.stderr.write("  %6d logical qubits before optimization\n" % len(self.all_bqm_variables()))

        # At the time of this writing, a BinaryQuadraticModel elides variables
        # with a weight of zero.  But then contract_variables complains that
        # the variable can't be found.  Hence, we add back all zero-valued
        # variables just to keep contract_variables from failing.  Similarly,
        # we ensure we have a strength defined (even if zero) for all chains to
        # prevent contract_variables from complaining.
        self.bqm.add_variables_from({q[0]: 0 for q in self.chains})
        self.bqm.add_variables_from({q[1]: 0 for q in self.chains})
        self.bqm.add_interactions_from({q: 0 for q in self.chains})

        # Remove variables that are made equivalent to other variable via
        # user-defined chains.
        order = self.traversal_order(self.chains)
        for u, v in order:
            self.bqm.contract_variables(u, v)
            self.contractions[v] = u

        # Say what we just did.
        if verbosity >= 2:
            sys.stderr.write("  %6d logical qubits after optimization\n\n" % len(self.all_bqm_variables(force_recompute=True)))

    def simplify_problem(self, verbosity):
        "Find and remove variables with known outputs."
        # Say what we're going to do.
        if verbosity >= 2:
            sys.stderr.write("Simplified the problem:\n\n")
            sys.stderr.write("  %6d logical qubits before optimization\n" % len(self.all_bqm_variables()))

        # Simplify the BQM.
        self.known_values = dimod.roof_duality.fix_variables(self.bqm, True)
        self.bqm.fix_variables(self.known_values)

        # Say what we just did.
        if verbosity >= 2:
            num_left = len(self.all_bqm_variables(force_recompute=True))
            sys.stderr.write("  %6d logical qubits after optimization\n\n" % num_left)
            if num_left == 0:
                sys.stderr.write("    Note: A complete solution can be found classically using roof duality.\n\n")

    def append_assertions_from_statements(self):
        "Convert user-specified chains, anti-chains, and pins to assertions."
        # Convert pending assertions to actual assertions.
        # TODO: Quote variables containing special characters.
        ap = AssertParser(self.qmasm)
        for s1, op, s2 in self.pending_asserts:
            ast = ap.parse(s1 + " " + op + " " + s2)
            ast.compile()
            self.assertions.append(ast)

    def merged_known_values(self):
        "Merge pinned, known_values, and chains into a s single dictionary."
        merged = {k: v for k, v in self.pinned}
        spin2bool = {-1: False, +1: True}
        for k, v in self.known_values.items():
            merged[k] = spin2bool[v]
        remaining = copy.copy(self.chains)
        while len(remaining) > 0:
            still_remaining = set()
            for q0, q1 in remaining:
                if q0 in merged:
                    merged[q1] = merged[q0]
                elif q1 in merged:
                    merged[q0] = merged[q1]
                else:
                    still_remaining.add((q0, q1))
            if len(still_remaining) == len(remaining):
                break  # No progress was made.
            remaining = still_remaining
        return merged

    def dangling_variables(self, num2syms):
        "Return a set of variables that are neither embedded nor have a known value."
        dangling = set()
        known_values = self.logical.merged_known_values()
        for i in range(len(num2syms)):
            if num2syms[i] == []:
                continue
            if i not in self.embedding and i not in known_values and i not in self.contractions:
                dangling.update(num2syms[i])
        return dangling

    def autoscale_coefficients(self, sampler):
        "Scale weights and strengths to match what the hardware allows."
        try:
            h_range = sampler.sampler.properties["h_range"]
            J_range = sampler.sampler.properties["j_range"]
            min_h = min([w for w in self.weights.values()])
            max_h = max([w for w in self.weights.values()])
            min_J = min([w for w in self.strengths.values()])
            max_J = max([w for w in self.strengths.values()])
            scale = min(abs(h_range[0]/min_h), abs(h_range[1]/max_h),
                        abs(J_range[0]/min_J), abs(J_range[1]/max_J))
            self.bqm.scale(scale)
            self.weights = self.bqm.linear
            self.strengths = self.bqm.quadratic
        except KeyError:
            pass  # Probably a software solver.

    def _output_embedding_verbosely(self, max_sym_name_len, num2syms):
        "Verbosely output the mapping from logical to physical qubits."
        sys.stderr.write("Established a mapping from logical to physical qubits:\n\n")
        sys.stderr.write("    Logical  %-*s  Physical\n" % (max_sym_name_len, "Name(s)"))
        sys.stderr.write("    -------  %s  -------------------------\n" % ("-" * max_sym_name_len))
        known_values = self.logical.merged_known_values()
        pin_map = {k: v for k, v in self.logical.pinned}
        for i in range(len(num2syms)):
            if num2syms[i] == []:
                continue
            name_list = " ".join(sorted(num2syms[i]))
            try:
                phys_list = " ".join(["%4d" % e for e in sorted(self.embedding[i])])
            except KeyError:
                try:
                    phys_list = "[Pinned to %s]" % repr(pin_map[i])
                except KeyError:
                    try:
                        phys_list = "[Provably %s]" % known_values[i]
                    except KeyError:
                        try:
                            phys_list = "[Same as logical %d]" % self.contractions[i]
                        except KeyError:
                            if self.embedding == {}:
                                phys_list = str(i)
                            else:
                                phys_list = "[Disconnected]"
            sys.stderr.write("    %7d  %-*s  %s\n" % (i, max_sym_name_len, name_list, phys_list))
        sys.stderr.write("\n")

    def _output_embedding_tersely(self, max_sym_name_len, num2syms):
        "Tersely output the mapping from logical to physical qubits."
        log2phys_comments = []
        known_values = self.logical.merged_known_values()
        pin_map = {k: v for k, v in self.logical.pinned}
        for i in range(len(num2syms)):
            if num2syms[i] == []:
                continue
            name_list = " ".join(sorted(num2syms[i]))
            try:
                phys_list = " ".join(["%4d" % e for e in sorted(self.embedding[i])])
            except KeyError:
                try:
                    phys_list = "[%s]" % repr(pin_map[i])
                except KeyError:
                    try:
                        phys_list = "[%s]" % known_values[i]
                    except KeyError:
                        try:
                            same = self.contractions[i]
                            same_str = " ".join(self.qmasm.sym_map.to_symbols(same))
                            phys_list = "[Same as %s]" % same_str
                        except KeyError:
                            if self.embedding == {}:
                                phys_list = str(i)
                            else:
                                phys_list = "[Disconnected]"
            log2phys_comments.append("# %s --> %s" % (name_list, phys_list))
        log2phys_comments.sort()
        sys.stderr.write("\n".join(log2phys_comments) + "\n")

    def output_embedding(self, verbosity, max_sym_name_len, num2syms):
        "Output the mapping from logical to physical qubits."
        if verbosity > 0:
            self._output_embedding_verbosely(max_sym_name_len, num2syms)
        else:
            self._output_embedding_tersely(max_sym_name_len, num2syms)

    def output_embedding_statistics(self):
        "Output various statistics about the embedding."
        # Tally the original set of variables.
        logical = self.logical
        all_vars = set(logical.weights)
        for q1, q2 in logical.strengths:
            all_vars.add(q1)
            all_vars.add(q2)

        # Output statistics.
        known_true = sum([v == 1 for v in logical.known_values.values()])
        known_false = len(logical.known_values) - known_true
        sys.stderr.write("Computed the following statistics of the logical-to-physical mapping:\n\n")
        sys.stderr.write("    Type      Metric           Value\n")
        sys.stderr.write("    --------  ---------------  -----\n")
        sys.stderr.write("    Logical   Variables        %5d\n" % len(all_vars))
        sys.stderr.write("    Logical     Provably true  %5d\n" % known_true)
        sys.stderr.write("    Logical     Provably false %5d\n" % known_false)
        sys.stderr.write("    Logical   Strengths        %5d\n" % len(logical.strengths))
        sys.stderr.write("    Logical     Equalities     %5d\n" % len(logical.chains))
        sys.stderr.write("    Logical     Inequalities   %5d\n" % len(logical.antichains))
        sys.stderr.write("    Physical  Spins            %5d\n" % len(self.weights))
        sys.stderr.write("    Physical  Couplers         %5d\n" % len(self.strengths))
        sys.stderr.write("\n")

        # Output some additional chain statistics.
        chain_lens = [len(c) for c in self.embedding.values()]
        max_chain_len = 0
        if chain_lens != []:
            max_chain_len = max(chain_lens)
        num_max_chains = len([l for l in chain_lens if l == max_chain_len])
        sys.stderr.write("    Maximum chain length = %d (occurrences = %d)\n\n" % (max_chain_len, num_max_chains))



###################################
# Parse an !assert directive      #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import qmasm
import re
import sys

class AST(object):
    "Represent an abstract syntax tree."

    def __init__(self, qmasm, type, value, kids=[]):
        self.qmasm = qmasm
        self.type = type
        self.value = value
        self.kids = kids
        self.code = lambda isb: qmasm.abend("Internal error: Attempt to evaluate an AST without compiling it first")     # Function that evaluates the AST given a mapping from identifiers to bits
        self._str = None   # Memoized string representation

    def _needs_parens(self):
        "Return True if an AST node should be parenthesized."
        return self.type == "factor" and self.kids[0].type == "conn"

    def _str_helper(self):
        "Do most of the work for the __str__ method."
        # Conditionally parenthesize all child strings.
        nkids = len(self.kids)
        kids_str = [str(k) for k in self.kids]
        for i in range(nkids):
            if self.kids[i]._needs_parens():
                kids_str[i] = "(" + kids_str[i] + ")"

        # Return ourself as a string.
        if nkids == 0:
            return str(self.value)
        if nkids == 1:
            if self.type == "unary" and self.value != "id":
                return "%s%s" % (self.value, kids_str[0])
            return kids_str[0]
        if nkids == 2:
            if self.value in ["*", "/", "%", "&", "<<", ">>", "**"]:
                return "%s%s%s" % (kids_str[0], self.value, kids_str[1])
            else:
                return "%s %s %s" % (kids_str[0], self.value, kids_str[1])
        if nkids == 3:
            if self.type == "if_expr":
                return "if %s then %s else %s endif" % (str(self.kids[0]), str(self.kids[1]), str(self.kids[2]))
        raise Exception("Internal error parsing (%s, %s)" % (repr(self.type), repr(self.value)))

    def __str__(self):
        if self._str == None:
            self._str = self._str_helper()
        return self._str

    def prefix_identifiers(self, prefix, next_prefix):
        "Prefix every identifier with a given string."
        if self.type == "ident":
            self.value = self.qmasm.apply_prefix(self.value, prefix, next_prefix)
        else:
            for k in self.kids:
                k.prefix_identifiers(prefix, next_prefix)

    def replace_ident(self, old_ident, new_ident):
        "Replace every occurrence of one identifer with another."
        if self.type == "ident":
            if self.value == old_ident:
                self.value = new_ident
        else:
            for k in self.kids:
                k.replace_ident(old_ident, new_ident)

    class EvaluationError(Exception):
        "Represent an exception thrown during AST evaluation."
        pass

    def _evaluate_ident(self, i2b):
        "Evaluate a variable."
        try:
            bit = i2b[self.value]
            if bit == None:
                raise self.EvaluationError("Unused variable %s" % self.value)
            return bit
        except KeyError:
            raise self.EvaluationError("Undefined variable %s" % self.value)

    def _compile_unary(self, kvals):
        "Compile a unary expression."
        if self.value == "-":
            return lambda i2b: -kvals[0](i2b)
        elif self.value == "~":
            return lambda i2b: ~kvals[0](i2b)
        elif self.value == "!":
            return lambda i2b: int(kvals[0](i2b) == 0)
        elif self.value in ["+", "id"]:
            return lambda i2b: kvals[0](i2b)
        else:
            raise self.EvaluationError('Internal error compiling unary "%s"' % self.value)

    def _evaluate_power(self, base, exp):
        "Raise one integer to the power of another."
        if exp < 0:
            raise self.EvaluationError("Negative powers (%d) are not allowed" % exp)
        return base**exp

    def _compile_arith(self, kvals):
        "Compile an arithmetic expression."
        if self.value == "+":
            return lambda i2b: kvals[0](i2b) + kvals[1](i2b)
        elif self.value == "-":
            return lambda i2b: kvals[0](i2b) - kvals[1](i2b)
        elif self.value == "*":
            return lambda i2b: kvals[0](i2b) * kvals[1](i2b)
        elif self.value == "/":
            return lambda i2b: kvals[0](i2b) // kvals[1](i2b)
        elif self.value == "%":
            return lambda i2b: kvals[0](i2b) % kvals[1](i2b)
        elif self.value == "&":
            return lambda i2b: kvals[0](i2b) & kvals[1](i2b)
        elif self.value == "|":
            return lambda i2b: kvals[0](i2b) | kvals[1](i2b)
        elif self.value == "^":
            return lambda i2b: kvals[0](i2b) ^ kvals[1](i2b)
        elif self.value == "<<":
            return lambda i2b: kvals[0](i2b) << kvals[1](i2b)
        elif self.value == ">>":
            return lambda i2b: kvals[0](i2b) >> kvals[1](i2b)
        elif self.value == "**":
            return lambda i2b: self._evaluate_power(kvals[0](i2b), kvals[1](i2b))
        else:
            raise self.EvaluationError("Internal error compiling arithmetic operator %s" % self.value)

    def _compile_rel(self, kvals):
        "Compile a relational expression."
        if self.value == "=":
            return lambda i2b: kvals[0](i2b) == kvals[1](i2b)
        elif self.value == "/=":
            return lambda i2b: kvals[0](i2b) != kvals[1](i2b)
        elif self.value == "<":
            return lambda i2b: kvals[0](i2b) < kvals[1](i2b)
        elif self.value == "<=":
            return lambda i2b: kvals[0](i2b) <= kvals[1](i2b)
        elif self.value == ">":
            return lambda i2b: kvals[0](i2b) > kvals[1](i2b)
        elif self.value == ">=":
            return lambda i2b: kvals[0](i2b) >= kvals[1](i2b)
        else:
            raise self.EvaluationError("Internal error compiling relational operator %s" % self.value)

    def _compile_conn(self, kvals):
        "Compile a logical connective."
        if self.value == "&&":
            return lambda i2b: kvals[0](i2b) and kvals[1](i2b)
        elif self.value == "||":
            return lambda i2b: kvals[0](i2b) or kvals[1](i2b)
        else:
            raise self.EvaluationError("Internal error compiling logical connective %s" % self.value)

    def _evaluate_if_expr(self, i2b, kvals):
        if kvals[0](i2b):
            return kvals[1](i2b)
        else:
            return kvals[2](i2b)

    def _compile_if_expr(self, kvals):
        "Compile an if...then...else expression."
        return lambda i2b: self._evaluate_if_expr(i2b, kvals)

    def _compile_node(self):
        """Compile the AST to a function that returns either True or False
        given a mapping from identifiers to bits."""
        kvals = [k._compile_node() for k in self.kids]
        if self.type == "ident":
            # Variable
            return lambda i2b: self._evaluate_ident(i2b)
        elif self.type == "int":
            # Constant
            return lambda i2b: self.value
        elif self.type == "unary":
            # Unary expression
            return self._compile_unary(kvals)
        elif len(kvals) == 1:
            # All other single-child nodes return their child unmodified.
            return kvals[0]
        elif self.type in ["power", "term", "expr"]:
            return self._compile_arith(kvals)
        elif self.type == "rel":
            return self._compile_rel(kvals)
        elif self.type == "conn":
            return self._compile_conn(kvals)
        elif self.type == "if_expr":
            return self._compile_if_expr(kvals)
        else:
            raise self.EvaluationError("Internal error compiling AST node of type %s, value %s" % (repr(self.type), repr(self.value)))

    def compile(self):
        "Compile an AST for faster evaluation."
        self.code = self._compile_node()

    def evaluate(self, i2b):
        "Evaluate the AST to a value, given a mapping from identifiers to bits."
        try:
            return self.code(i2b)
        except self.EvaluationError as e:
            self.qmasm.abend("%s in assertion %s" % (e, self))

class AssertParser(object):
    int_re = re.compile(r'\d+')
    conn_re = re.compile(r'\|\||&&')
    rel_re = re.compile(r'/?=|[<>]=?')
    arith_re = re.compile(r'[-+/%&\|^~!]|>>|<<|\*\*?')
    keyword_re = re.compile(r'\b(if|then|else|endif)\b')

    def __init__(self, qmasm):
        self.qmasm = qmasm
    
    class ParseError(Exception):
        pass

    def lex(self, s):
        "Split a string into tokens (tuples of type and value)."
        tokens = []
        s = s.lstrip()
        while len(s) > 0:
            # Match parentheses.
            if s[0] == "(":
                tokens.append(("lparen", "("))
                s = s[1:].lstrip()
                continue
            if s[0] == ")":
                tokens.append(("rparen", ")"))
                s = s[1:].lstrip()
                continue

            # Match keywords.
            mo = self.keyword_re.match(s)
            if mo != None:
                match = mo.group(0)
                tokens.append((match, match))
                s = s[len(match):].lstrip()
                continue

            # Match positive integers.
            mo = self.int_re.match(s)
            if mo != None:
                match = mo.group(0)
                tokens.append(("int", int(match)))
                s = s[len(match):].lstrip()
                continue

            # Match connectives.
            mo = self.conn_re.match(s)
            if mo != None:
                match = mo.group(0)
                tokens.append(("conn", match))
                s = s[len(match):].lstrip()
                continue

            # Match "<<" and ">>" before we match "<" and ">".
            if len(s) >= 2 and (s[:2] == "<<" or s[:2] == ">>"):
                tokens.append(("arith", s[:2]))
                s = s[2:].lstrip()
                continue

            # Match relational operators.
            mo = self.rel_re.match(s)
            if mo != None:
                match = mo.group(0)
                tokens.append(("rel", match))
                s = s[len(match):].lstrip()
                continue

            # Match "**" before we match "*".
            if len(s) >= 2 and s[:2] == "**":
                tokens.append(("power", s[:2]))
                s = s[2:].lstrip()
                continue

            # Match arithmetic operators.
            mo = self.arith_re.match(s)
            if mo != None:
                match = mo.group(0)
                tokens.append(("arith", match))
                s = s[len(match):].lstrip()
                continue

            # Everything else is an identifier.
            mo = self.qmasm.ident_re.match(s)
            if mo != None:
                match = mo.group(0)
                tokens.append(("ident", match))
                s = s[len(match):].lstrip()
                continue
            raise self.ParseError("Failed to parse %s" % s)
        tokens.append(("EOF", "EOF"))
        return tokens

    def advance(self):
        "Advance to the next symbol."
        self.tokidx += 1
        self.sym = self.tokens[self.tokidx]

    def accept(self, ty):
        """Advance to the next token if the current token matches a given
        token type and return True.  Otherwise, return False."""
        if self.sym[0] == ty:
            self.advance()
            return True
        return False

    def expect(self, ty):
        """Advance to the next token if the current token matches a given
        token.  Otherwise, fail."""
        if not self.accept(ty):
            raise self.ParseError("Expected %s but saw %s" % (ty, repr(self.sym[1])))

    def generic_operator(self, return_type, child_method, sym_type, valid_ops):
        "Match one or more somethings to produce something else."
        # Produce a list of ASTs representing children.
        c = child_method()
        ops = [self.sym[1]]
        asts = [c]
        while self.sym[0] == sym_type and ops[-1] in valid_ops:
            self.advance()
            c = child_method()
            ops.append(self.sym[1])
            asts.append(c)

        # Handle the trivial case of the identity operation.
        if len(asts) == 1:
            return AST(self.qmasm, return_type, None, asts)

        # Merge the ASTs in a left-associative fashion into a single AST.
        ops.pop()
        while len(asts) > 1:
            asts = [AST(self.qmasm, return_type, ops[0], [asts[0], asts[1]])] + asts[2:]
            ops.pop(0)
        return asts[0]

    def if_expr(self):
        "Return an if...then...else expression."
        self.expect("if")
        cond = self.conjunction()
        self.expect("then")
        then_expr = self.expression()
        self.expect("else")
        else_expr = self.expression()
        self.expect("endif")
        return AST(self.qmasm, "if_expr", None, [cond, then_expr, else_expr])

    def factor(self):
        "Return a factor (variable, integer, or expression)."
        val = self.sym[1]
        if self.accept("ident"):
            child = AST(self.qmasm, "ident", val)
        elif self.accept("int"):
            child = AST(self.qmasm, "int", val)
        elif self.accept("lparen"):
            child = self.disjunction()
            self.expect("rparen")
        elif self.sym[0] == "arith":
            child = self.unary()
        elif self.sym[0] == "if":
            child = self.if_expr()
        elif val == "EOF":
            raise self.ParseError("Parse error at end of expression")
        else:
            raise self.ParseError('Parse error at "%s"' % val)
        return AST(self.qmasm, "factor", None, [child])

    def power(self):
        "Return a factor or a factor raised to the power of a second factor."
        f1 = self.factor()
        op = self.sym[1]
        if self.sym[0] == "power" and op == "**":
            self.advance()
            f2 = self.power()
            return AST(self.qmasm, "power", op, [f1, f2])
        return AST(self.qmasm, "power", None, [f1])

    def unary(self):
        "Return a unary operator applied to a power."
        op = self.sym[1]
        if op in ["+", "-", "~", "!"]:
            self.advance()
        else:
            op = "id"
        return AST(self.qmasm, "unary", op, [self.power()])

    def term(self):
        "Return a term (product of one or more unaries)."
        return self.generic_operator("term", self.unary, "arith", ["*", "/", "%", "&", "<<", ">>"])

    def expression(self):
        "Return an expression (sum of one or more terms)."
        return self.generic_operator("expr", self.term, "arith", ["+", "-", "|", "^"])

    def comparison(self):
        "Return a comparison of exactly two expressions."
        e1 = self.expression()
        op = self.sym[1]
        if self.sym[0] != "rel":
            return AST(self.qmasm, "rel", None, [e1])
        self.advance()
        e2 = self.expression()
        return AST(self.qmasm, "rel", op, [e1, e2])

    def conjunction(self):
        "Return a conjunction (logical AND of one or more comparisons)."
        return self.generic_operator("conn", self.comparison, "conn", ["&&"])

    def disjunction(self):
        "Return a disjunction (logical OR of one or more conjunctions)."
        return self.generic_operator("conn", self.conjunction, "conn", ["||"])

    def parse(self, s):
        "Parse a relational expression into an AST"
        self.tokens = self.lex(s)
        self.tokidx = -1
        self.advance()
        try:
            ast = self.disjunction()
            if self.sym[0] != "EOF":
                raise self.ParseError('Parse error at "%s"' % self.sym[1])
        except self.ParseError as e:
            qmasm.abend('%s in "%s"' % (e, s))
        return ast



###################################
# Parse the QMASM command line    #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import argparse
import shlex
import string
import sys

class ParseCommandLine(object):
    def parse_command_line(self):
        "Parse the QMASM command line.  Return an argparse.Namespace."

        # Define all of our command-line arguments.
        cl_parser = argparse.ArgumentParser(description="Assemble a symbolic Hamiltonian into a numeric one")
        cl_parser.add_argument("input", nargs="*",
                               help="file(s) from which to read a symbolic Hamiltonian")
        cl_parser.add_argument("-v", "--verbose", action="count", default=0,
                               help="increase output verbosity (can be specified repeatedly)")
        cl_parser.add_argument("--run", action="store_true",
                               help="run the program on the current solver")
        cl_parser.add_argument("-o", "--output", metavar="FILE", default="<stdout>",
                               help="file to which to write weights and strengths (default: none)")
        cl_parser.add_argument("-O", type=int, nargs="?", const=1, default=0,
                               metavar="LEVEL",
                               help="optimize the layout; at -O1, remove unnecessary qubits")
        cl_parser.add_argument("--pin", action="append",
                               help="pin a set of qubits to a set of true or false values")
        cl_parser.add_argument("--format", choices=["qubist", "qbsolv", "qmasm", "minizinc", "bqpjson"], default="qubist",
                               help="output-file format")
        cl_parser.add_argument("--values", choices=["bools", "ints"], default="bools",
                               help="output solution values as Booleans or integers (default: bools)")
        cl_parser.add_argument("--profile", type=str, default=None, metavar="NAME",
                               help="Profile name from dwave.conf to use")
        cl_parser.add_argument("--solver", type=str, default=None, metavar="NAME",
                               help='Solver name from dwave.conf to use or one of the special names "exact", "sim-anneal", or "tabu"')
        cl_parser.add_argument("--chain-strength", metavar="NEG_NUM", type=float,
                               help="negative-valued chain strength (default: automatic)")
        cl_parser.add_argument("--pin-weight", metavar="NEG_NUM", type=float,
                               help="negative-valued pin weight (default: automatic)")
        cl_parser.add_argument("--qubo", action="store_true",
                               help="treat inputs as QUBOs rather than Ising systems")
        cl_parser.add_argument("--samples", metavar="POS_INT", type=int, default=1000,
                               help="number of samples to take (default: 1000)")
        cl_parser.add_argument("--anneal-time", metavar="POS_INT", type=int, default=None,
                               help="annealing time in microseconds (default: automatic)")
        cl_parser.add_argument("--spin-revs", metavar="POS_INT", type=int, default=0,
                               help="number of spin-reversal transforms to perform (default: 0)")
        cl_parser.add_argument("--topology-file", default=None, metavar="FILE",
                               help="name of a file describing the topology (list of vertex pairs)")
        cl_parser.add_argument("--postproc", choices=["none", "sample", "opt"],
                               default="none",
                               help='type of postprocessing to perform (default: "none")')
        cl_parser.add_argument("--show", choices=["valid", "all", "best"], default="valid",
                               help='show valid solutions, all solutions, or the best (even if invalid) solutions (default: "valid")')
        cl_parser.add_argument("--always-embed", action="store_true",
                               help="when writing an output file, embed the problem in the physical topology even when not required (default: false)")
        cl_parser.add_argument("--qbsolv", action="store_true",
                               help="wrap the solver with QBSolv to break up large problems into pieces")

        # Parse the command line.
        cl_args = cl_parser.parse_args()

        # Perform a few sanity checks on the parameters.
        if cl_args.chain_strength != None and cl_args.chain_strength >= 0.0:
            self.warn("A non-negative chain strength (%.20g) was specified\n" % cl_args.chain_strength)
        if cl_args.pin_weight != None and cl_args.pin_weight >= 0.0:
            self.warn("A non-negative pin strength (%.20g) was specified\n" % cl_args.pin_weight)
        if cl_args.spin_revs > cl_args.samples:
            self.abend("The number of spin reversals is not allowed to exceed the number of samples")
        return cl_args

    def get_command_line(self):
        "Return the command line as a string, properly quoted."
        return " ".join([shlex.quote(a) for a in sys.argv])

    def report_command_line(self, cl_args):
        "For provenance and debugging purposes, report our command line parameters."
        # Output the command line as is.
        verbosity = cl_args.verbose
        if verbosity < 1:
            return
        sys.stderr.write("Command line provided:\n\n")
        sys.stderr.write("    %s\n\n" % self.get_command_line())

        # At higher levels of verbosity, output every single option.
        if verbosity < 2:
            return
        sys.stderr.write("All QMASM parameters:\n\n")
        params = vars(cl_args)
        klen = max([len(a) for a in params.keys()])
        klen = max(klen + 2, len("Option"))   # +2 for "--"
        vlen = max([len(repr(a)) for a in params.values()])
        klen = max(vlen, len("Value(s)"))
        sys.stderr.write("    %-*s  %-*s\n" % (klen, "Option", vlen, "Value(s)"))
        sys.stderr.write("    %s  %s\n" % ("-" * klen, "-" * vlen))
        for k in sorted(params.keys()):
            kname = k.replace("_", "-")
            sys.stderr.write("    %-*s  %-*s\n" % (klen, "--" + kname, vlen, repr(params[k])))
        sys.stderr.write("\n")



########################################
# Store and filter solutions for QMASM #
# By Scott Pakin <pakin@lanl.gov>      #
########################################

import copy
import math
import numpy as np
import re
import sys
from dwave.embedding import unembed_sampleset, chain_breaks, chain_break_frequency
from scipy.stats import median_absolute_deviation

# Define a class to represent a single solution.
class Solution:
    "Represent a near-minimal state of a spin system."

    def __init__(self, problem, sym2col, num2col, raw_ss, fixed_soln, tally, energy, all_vars):
        # Map named variables to spins.
        self.problem = problem
        self.sym2col = sym2col
        self.num2col = num2col
        self.raw_ss = raw_ss
        self.fixed_soln = fixed_soln
        self.tally = tally
        self.energy = energy
        self.id = 0                   # Map from spins to an int
        self._checked_asserts = None  # Memoized result of check_assertions

        # Compute an ID for the solution.
        for s, c in sorted(self.sym2col.items()):
            if "$" in s and not all_vars:
                continue
            self.id = self.id*2 + int((self.fixed_soln[c] + 1)//2)

        # Ensure every symbol has an associated value.
        self.sym2bool = self._all_symbol_values()

    def _all_symbol_values(self):
        """Return a mapping from every symbol to a value, including symbols
        with no physical representation."""
        # Assign values to all symbols that were eventually embedded.
        sym2bool = {}
        sym_map = self.problem.qmasm.sym_map
        all_sym2num = sym_map.symbol_number_items()
        for s, n in all_sym2num:
            try:
                c = self.num2col[n]
                sym2bool[s] = self.fixed_soln[c] == 1
            except KeyError:
                pass  # We handle non-embedded symbols below.

        # Include values inferred from roof duality.
        for s, spin in self.problem.logical.known_values.items():
            sym2bool[s] = spin == 1

        # Include variables pinned explicitly by the program or user.
        for num, bval in self.problem.logical.pinned:
            for s in sym_map.to_symbols(num):
                sym2bool[s] = bval

        # Include contracted variables (those chained to other another
        # variable's value).
        contracted = self.problem.traversal_order(self.problem.logical.chains)
        contracted.reverse()
        for num1, num2 in contracted:
            for s1 in sym_map.to_symbols(num1):
                for s2 in sym_map.to_symbols(num2):
                    try:
                        sym2bool[s2] = sym2bool[s1]
                    except KeyError:
                        pass
        return sym2bool

    def broken_chains(self):
        "Return True if the solution contains broken chains."
        unbroken = unembed_sampleset(self.raw_ss, self.problem.embedding,
                                     self.problem.logical.bqm,
                                     chain_break_method=chain_breaks.discard)
        return len(unbroken) == 0

    def broken_user_chains(self):
        "Return True if the solution contains broken user-specified chains or anti-chains."
        # Compare logical qubits for equal values.
        for lq1, lq2 in self.problem.logical.chains:
            try:
                idx1, idx2 = self.num2col[lq1], self.num2col[lq2]
                if self.fixed_soln[idx1] != self.fixed_soln[idx2]:
                    return True
            except KeyError:
                pass  # Elided logical qubit
        for lq1, lq2 in self.problem.logical.antichains:
            try:
                idx1, idx2 = self.num2col[lq1], self.num2col[lq2]
                if self.fixed_soln[idx1] == self.fixed_soln[idx2]:
                    return True
            except KeyError:
                pass  # Elided logical qubit
        return False

    def broken_pins(self):
        "Return True if the solution contains broken pins."
        bool2spin = [-1, +1]
        for pq, pin in self.problem.logical.pinned:
            try:
                idx = self.num2col[pq]
                if self.fixed_soln[idx] != bool2spin[pin]:
                    return True
            except KeyError:
                pass  # Elided logical qubit
        return False

    def check_assertions(self, stop_on_fail=False):
        "Return the result of applying each assertion."
        # Return the previous result, if any.
        if self._checked_asserts != None:
            return self._checked_asserts

        # Test each assertion in turn.
        results = []
        sym2bit = {s: int(bval) for s, bval in self.sym2bool.items()}
        for a in self.problem.assertions:
            results.append((str(a), a.evaluate(sym2bit)))
            if stop_on_fail and not results[-1][1]:
                return results
        self._checked_asserts = results
        return self._checked_asserts

    def failed_assertions(self, stop_on_fail):
        "Return True if the solution contains failed assertions."
        return not all([a[1] for a in self.check_assertions(stop_on_fail)])

    def output_asserts(self, verbosity):
        "Helper function for output_solution that outputs assertion results."
        # Do nothing if the program contains no assertions.  Otherwise, output
        # some header text.
        if len(self.problem.assertions) == 0:
            return
        if verbosity >= 2:
            print("    Assertions made")
            print("    ---------------")
        else:
            print("    Assertions failed")
            print("    -----------------")

        # Output each assertion in turn.
        n_failed = 0
        for (astr, ok) in self.check_assertions():
            if not ok:
                print("    FAIL: %s" % astr)
                n_failed += 1
            elif verbosity >= 2:
                print("    PASS: %s" % astr)
        if verbosity < 2 and n_failed == 0:
            print("    [none]")
        print("")

    def _numeric_solution(self):
        "Convert single- and multi-bit values to numbers."
        # Map each name to a number and to the number of bits required.
        idx_re = re.compile(r'^([^\[\]]+)\[(\d+)\]$')
        name2num = {}
        name2nbits = {}
        for nm, bval in self.sym2bool.items():
            # Parse the name into a prefix and array index.
            match = idx_re.search(nm)
            if match == None:
                # No array index: Treat as a 1-bit number.
                name2num[nm] = int(bval)
                name2nbits[nm] = 1
                continue

            # Integrate the current spin into the overall number.
            array, idx = match.groups()
            b = int(bval) << int(idx)
            try:
                name2num[array] += b
                name2nbits[array] = max(name2nbits[array], int(idx) + 1)
            except KeyError:
                name2num[array] = b
                name2nbits[array] = int(idx) + 1

        # Merge the two maps.
        return {nm: (name2num[nm], name2nbits[nm]) for nm in name2num.keys()}

    def output_solution_bools(self, verbosity):
        "Output the solution as a list of Boolean variables."
        # Find the width of the longest symbol name.
        all_syms = self.sym2bool.keys()
        if verbosity < 2:
            all_syms = [s for s in all_syms if "$" not in s]
        all_syms = sorted(all_syms)
        max_sym_name_len = max(8, max([len(s) for s in all_syms]))

        # Output one symbol per line.
        print("    %-*s  Value" % (max_sym_name_len, "Variable"))
        print("    %s  -----" % ("-" * max_sym_name_len))
        for s in all_syms:
            bval = self.sym2bool[s]
            print("    %-*s  %s" % (max_sym_name_len, s, bval))
        print("")

    def output_solution_ints(self, verbosity):
        "Output the solution as integer values."
        # Convert each value to a decimal and a binary string.  Along the way,
        # find the width of the longest name and the largest number.
        name2info = self._numeric_solution()
        if verbosity < 2:
            name2info = {n: i for n, i in name2info.items() if "$" not in n}
        max_sym_name_len = max([len(s) for s in list(name2info.keys()) + ["Name"]])
        max_decimal_len = 7
        max_binary_len = 6
        name2strs = {}
        for name, info in name2info.items():
            bstr = ("{0:0" + str(info[1]) + "b}").format(info[0])
            dstr = str(info[0])
            max_binary_len = max(max_binary_len, len(bstr))
            max_decimal_len = max(max_decimal_len, len(dstr))
            name2strs[name] = (bstr, dstr)

        # Output one name per line.
        print("    %-*s  %-*s  Decimal" % (max_sym_name_len, "Name", max_binary_len, "Binary"))
        print("    %s  %s  %s" % ("-" * max_sym_name_len, "-" * max_binary_len, "-" * max_decimal_len))
        for name, (bstr, dstr) in sorted(name2strs.items()):
            print("    %-*s  %*s  %*s" % (max_sym_name_len, name, max_binary_len, bstr, max_decimal_len, dstr))
        print("")

class Solutions(object):
    "Represent all near-minimal states of a spin system."

    def __init__(self, answer, problem, all_vars):
        # Store our arguments.
        self.answer = answer
        self.problem = problem

        # Unembed the solutions.  Fix rather than discard invalid solutions.
        if self.problem.embedding == {}:
            fixed_answer = self.answer.copy()
        else:
            fixed_answer = unembed_sampleset(self.answer, self.problem.embedding,
                                             self.problem.logical.bqm,
                                             chain_break_method=chain_breaks.majority_vote)

        # Construct one Solution object per solution.
        self.solutions = []
        energies = fixed_answer.record.energy
        tallies = fixed_answer.record.num_occurrences
        fixed_solns = fixed_answer.record.sample
        raw_solns = self.answer.record.sample
        sym2col, num2col = self._map_to_column(fixed_answer,
                                               self.problem.embedding,
                                               self.problem.qmasm.sym_map)
        for i in range(len(fixed_solns)):
            sset = self.answer.slice(i, i + 1)
            self.solutions.append(Solution(self.problem, sym2col, num2col, sset,
                                           fixed_solns[i], tallies[i], energies[i],
                                           all_vars))

        # Store the frequency of chain breaks across the entire SampleSet.
        self.chain_breaks = chain_break_frequency(self.answer, self.problem.embedding)

    def _map_to_column(self, sset, embedding, sym_map):
        """Return a mapping from a symbol name and from a logical qubit number
        to a column number in a SampleSet."""
        # Compute a mapping from logical qubit to column number.
        num2col = {}
        for i in range(len(sset.variables)):
            num2col[sset.variables[i]] = i

        # Compute a mapping from symbol to a column number.  We need to go
        # symbol --> logical qubit --> column number.
        sym2col = {}
        for sym, num in sym_map.symbol_number_items():
            try:
                sym2col[sym] = num2col[num]
            except KeyError:
                pass
        return sym2col, num2col

    def report_timing_information(self, verbosity):
        "Output solver timing information."
        if verbosity == 0:
            return
        timing_info = sorted(list(self.answer.info["timing"].items()))
        sys.stderr.write("Timing information:\n\n")
        sys.stderr.write("    %-30s %-10s\n" % ("Measurement", "Value (us)"))
        sys.stderr.write("    %s %s\n" % ("-" * 30, "-" * 10))
        for timing_value in timing_info:
            sys.stderr.write("    %-30s %10d\n" % timing_value)
        sys.stderr.write("\n")

    def report_chain_break_information(self, verbosity):
        "Output information about common chain breaks."
        # Ensure we have something to report.
        if verbosity < 2:
            return
        sys.stderr.write("Chain-break frequencies:\n\n")
        total_breakage = sum(self.chain_breaks.values())
        if total_breakage == 0.0:
            sys.stderr.write("    [No broken chains encountered]\n\n")
            return

        # Report only chains that have ever been broken.
        chain_breaks = []
        max_name_len = 11
        sym_map = self.problem.qmasm.sym_map
        for n, f in self.chain_breaks.items():
            ss = sym_map.to_symbols(n)
            sstr = " ".join(sorted(ss))
            chain_breaks.append((sstr, f))
            max_name_len = max(max_name_len, len(sstr))
        chain_breaks.sort()
        sys.stderr.write("    %-*s  Broken\n" % (max_name_len, "Variable(s)"))
        sys.stderr.write("    %s  -------\n" % ("-" * max_name_len))
        for sstr, f in chain_breaks:
            if f > 0.0:
                sys.stderr.write("    %-*s  %6.2f%%\n" % (max_name_len, sstr, f*100.0))
        sys.stderr.write("\n")

    def report_energy_tallies(self, filtered_solns, full_output):
        "Output information about the observed energy levels."
        # Acquire a tally of each energy level.
        energies_tallies = list(zip(self.answer.record.energy, self.answer.record.num_occurrences))

        # If the caller requested full output, generate a complete energy
        # histogram.
        if full_output:
            # Merge energies based on a fixed number of digits after
            # the decimal point.
            raw_hist = {}
            for e, t in energies_tallies:
                estr = "%11.4f" % e
                try:
                    raw_hist[estr] += t
                except KeyError:
                    raw_hist[estr] = t

            # Output a histogram of energy values.
            sys.stderr.write("Histogram of raw energy values (merged to 4 digits after the decimal point):\n\n")
            sys.stderr.write("    Energy       Tally\n")
            sys.stderr.write("    -----------  --------\n")
            for et in sorted(raw_hist.items(), reverse=True):
                sys.stderr.write("    %s  %8d\n" % et)
            sys.stderr.write("\n")

        # Compute various statistics on the raw energies.
        raw_energies = np.array([v
                                 for lst in [[e]*t for e, t in energies_tallies]
                                 for v in lst])
        raw_min = np.amin(raw_energies)
        raw_mean = np.mean(raw_energies)
        raw_median = np.median(raw_energies)
        raw_mad = median_absolute_deviation(raw_energies)
        raw_stddev = np.std(raw_energies)
        raw_max = np.amax(raw_energies)

        # Compute various statistics on the filtered energies.
        filtered_energies = np.array([v
                                      for lst in [[s.energy]*s.tally for s in filtered_solns.solutions]
                                      for v in lst])
        if len(filtered_energies) == 0:
            filtered_min = math.nan
            filtered_mean = math.nan
            filtered_median = math.nan
            filtered_mad = math.nan
            filtered_stddev = math.nan
            filtered_max = math.nan
        else:
            filtered_min = np.amin(filtered_energies)
            filtered_mean = np.mean(filtered_energies)
            filtered_median = np.median(filtered_energies)
            filtered_mad = median_absolute_deviation(filtered_energies)
            filtered_stddev = np.std(filtered_energies)
            filtered_max = np.amax(filtered_energies)

        # Output energy statistics.
        sys.stderr.write("Energy statistics:\n\n")
        sys.stderr.write("    Statistic  All solutions           Filtered solutions\n")
        sys.stderr.write("    ---------  ----------------------  -----------------------\n")
        sys.stderr.write("    Minimum    %11.4f              %11.4f\n" % (raw_min, filtered_min))
        sys.stderr.write("    Median     %11.4f +/- %-7.4f  %11.4f +/- %-7.4f\n" % (raw_median, raw_mad, filtered_median, filtered_mad))
        sys.stderr.write("    Mean       %11.4f +/- %-7.4f  %11.4f +/- %-7.4f\n" % (raw_mean, raw_stddev, filtered_mean, filtered_stddev))
        sys.stderr.write("    Maximum    %11.4f              %11.4f\n" % (raw_max, filtered_max))
        sys.stderr.write("\n")

    def discard_broken_chains(self):
        "Discard solutions with broken chains.  Return the remaining solutions."
        if self.problem.embedding == {}:
            return self.solutions
        return [s for s in self.solutions if not s.broken_chains()]

    def discard_broken_user_chains(self):
        "Discard solutions with broken user-specified chains.  Return the remaining solutions."
        return [s for s in self.solutions if not s.broken_user_chains()]

    def discard_broken_pins(self):
        "Discard solutions with broken pins.  Return the remaining solutions."
        return [s for s in self.solutions if not s.broken_pins()]

    def discard_failed_assertions(self, stop_on_fail):
        "Discard solutions with failed assertions.  Return the remaining solutions."
        return [s for s in self.solutions if not s.failed_assertions(stop_on_fail)]

    def discard_non_minimal(self):
        "Discard solutions with non-minimal energy.  Return the remaining solutions."
        if len(self.solutions) == 0:
            return []
        min_energy = min([s.energy for s in self.solutions])
        return [s for s in self.solutions if s.energy == min_energy]

    def merge_duplicates(self):
        """Merge duplicate solutions (same assignments to all non-ignored
        variables).  Replace the merged solutions."""
        id2soln = {}
        for s in copy.deepcopy(self.solutions):  # Deep copy because of destructive update of tally
            try:
                id2soln[s.id].tally += s.tally
            except KeyError:
                id2soln[s.id] = s
        solutions = list(id2soln.values())
        solutions.sort(key=lambda s: (s.energy, s.id))
        return solutions

    def filter(self, show, verbose, nsamples):
        '''Return solutions as filtered according to the "show" parameter.
        Output information about the filtering based on the "verbose"
        parameter.'''
        # Prepare various views of the solutions.
        all_solns = copy.copy(self)
        valid_solns = copy.copy(self)
        best_solns = copy.copy(self)

        # Output the total number and number of unique solutions.
        if verbose >= 1:
            ndigits = len(str(nsamples))
            sys.stderr.write("Number of solutions found (filtered cumulatively):\n\n")
            sys.stderr.write("    %*d total solutions\n" % (ndigits, nsamples))
            sys.stderr.write("    %*d unique solutions\n" % (ndigits, len(all_solns.solutions)))

        # Filter out broken chains.
        if verbose >= 1 or show == "valid":
            valid_solns.solutions = valid_solns.discard_broken_chains()
            if verbose >= 1:
                sys.stderr.write("    %*d with no broken chains\n" % (ndigits, len(valid_solns.solutions)))
        if show == "best":
            filtered_best_solns = best_solns.discard_broken_chains()
            if len(filtered_best_solns) > 1:
                best_solns.solutions = filtered_best_solns

        # Filter out broken user-defined chains.
        if verbose >= 1 or show == "valid":
            valid_solns.solutions = valid_solns.discard_broken_user_chains()
            if verbose >= 1:
                sys.stderr.write("    %*d with no broken user-specified chains or anti-chains\n" % (ndigits, len(valid_solns.solutions)))
        if show == "best":
            filtered_best_solns = best_solns.discard_broken_user_chains()
            if len(filtered_best_solns) > 1:
                best_solns.solutions = filtered_best_solns

        # Filter out broken pins.
        if verbose >= 1 or show == "valid":
            valid_solns.solutions = valid_solns.discard_broken_pins()
            if verbose >= 1:
                sys.stderr.write("    %*d with no broken pins\n" % (ndigits, len(valid_solns.solutions)))
        if show == "best":
            filtered_best_solns = best_solns.discard_broken_pins()
            if len(filtered_best_solns) > 1:
                best_solns.solutions = filtered_best_solns

        # Filter out failed assertions.
        stop_on_fail = show == "valid" and verbose < 2
        if verbose >= 1 or show == "valid":
            valid_solns.solutions = valid_solns.discard_failed_assertions(stop_on_fail)
            if verbose >= 1:
                sys.stderr.write("    %*d with no failed assertions\n" % (ndigits, len(valid_solns.solutions)))
        if show == "best":
            filtered_best_solns = best_solns.discard_failed_assertions(stop_on_fail)
            if len(filtered_best_solns) > 1:
                best_solns.solutions = filtered_best_solns

        # Filter out solutions that are not at minimal energy.
        if verbose >= 1 or show == "valid":
            valid_solns.solutions = valid_solns.discard_non_minimal()
            if verbose >= 1:
                sys.stderr.write("    %*d at minimal energy\n" % (ndigits, len(valid_solns.solutions)))
        if show == "best":
            filtered_best_solns = best_solns.discard_non_minimal()
            if len(filtered_best_solns) > 1:
                best_solns.solutions = filtered_best_solns

        # Merge duplicate solutions.
        if verbose >= 1 or show == "valid":
            valid_solns.solutions = valid_solns.merge_duplicates()
            if verbose >= 1:
                sys.stderr.write("    %*d excluding duplicate variable assignments\n" % (ndigits, len(valid_solns.solutions)))
                sys.stderr.write("\n")
        if show == "best":
            filtered_best_solns = best_solns.merge_duplicates()
            if len(filtered_best_solns) > 1:
                best_solns.solutions = filtered_best_solns

        # Return the requested set of solutions.  Sort these first by energy
        # then by ID.
        soln_key = lambda s: (s.energy, s.id)
        if show == "valid":
            valid_solns.solutions.sort(key=soln_key)
            return valid_solns
        elif show == "all":
            all_solns.solutions.sort(key=soln_key)
            return all_solns
        elif show == "best":
            best_solns.solutions.sort(key=soln_key)
            return best_solns
        else:
            raise Exception("Internal error processing --show=%s" % show)

    def output_solutions(self, style, verbosity, show_asserts):
        "Output a user-readable solution to the standard output device."
        # Output each solution in turn.
        if len(self.solutions) == 0:
            print("No valid solutions found.")
            return
        for snum in range(len(self.solutions)):
            soln = self.solutions[snum]
            print("Solution #%d (energy = %.4f, tally = %s):\n" % (snum + 1, soln.energy, soln.tally))
            if style == "bools":
                soln.output_solution_bools(verbosity)
            elif style == "ints":
                soln.output_solution_ints(verbosity)
            else:
                raise Exception('Output style "%s" not recognized' % style)
            if show_asserts:
                soln.output_asserts(verbosity)



###################################
# Output QUBOs in various formats #
# By Scott Pakin <pakin@lanl.gov> #
###################################

import datetime
import json
import random
import shlex
import sys

class OutputMixin(object):
    "Provide functions for outputting problems and solutions."

    def open_output_file(self, oname):
        "Open a file or standard output."
        if oname == "<stdout>":
            outfile = sys.stdout
        else:
            try:
                outfile = open(oname, "w")
            except IOError:
                self.abend('Failed to open %s for output' % oname)
        return outfile

    def output_qmasm(self, outfile):
        "Output weights and strengths as a flattened QMASM source file."
        for p in self.program:
            outfile.write("%s\n" % p.as_str())

    def output_bqpjson(self, outfile, as_qubo, problem):
        "Output weights and strengths in bqpjson format, either Ising or QUBO."
        # Prepare the "easy" fields.
        bqp = {}
        bqp["version"] = "1.0.0"
        bqp["id"] = random.randint(2**20, 2**60)
        bqp["scale"] = 1.0
        bqp["offset"] = 0.0
        if as_qubo:
            bqp["variable_domain"] = "boolean"
        else:
            bqp["variable_domain"] = "spin"

        # Prepare the list of all variables.
        var_ids = set(problem.weights.keys())
        for q1, q2 in problem.strengths.keys():
            var_ids.add(q1)
            var_ids.add(q2)
        bqp["variable_ids"] = sorted(var_ids)

        # Prepare the linear terms.
        lin_terms = []
        for q, wt in sorted(problem.weights.items()):
            lin_terms.append({
                "id": q,
                "coeff": wt})
        bqp["linear_terms"] = lin_terms

        # Prepare the quadratic terms.
        quad_terms = []
        strengths = self.canonicalize_strengths(problem.strengths)
        for (q1, q2), wt in sorted(strengths.items()):
            quad_terms.append({
                "id_tail": q1,
                "id_head": q2,
                "coeff": wt})
        bqp["quadratic_terms"] = quad_terms

        # Prepare some metadata.
        metadata = {}
        if as_qubo:
            metadata["description"] = "QUBO problem compiled by QMASM (https://github.com/lanl/qmasm)"
        else:
            metadata["description"] = "Ising problem compiled by QMASM (https://github.com/lanl/qmasm)"
        metadata["command_line"] = self.get_command_line()
        metadata["generated"] = datetime.datetime.utcnow().isoformat()
        metadata["variable_names"] = {s: [n]
                                      for s, n in self.sym_map.symbol_number_items()}
        bqp["metadata"] = metadata

        # Output the problem in JSON format.
        outfile.write(json.dumps(bqp, indent=2, sort_keys=True) + "\n")

    def output_minizinc(self, outfile, problem, energy=None):
        "Output weights and strengths as a MiniZinc constraint problem."
        # Write some header information.
        outfile.write("""% Use MiniZinc to minimize a given Hamiltonian.
    %
    % Producer:     QMASM (https://github.com/lanl/qmasm/)
    % Author:       Scott Pakin (pakin@lanl.gov)
    """)
        outfile.write("%% Command line: %s\n\n" % " ".join([shlex.quote(a) for a in sys.argv]))

        # The model is easier to express as a QUBO so convert to that format.
        if problem.qubo:
            qprob = problem
        else:
            qprob = problem.convert_to_qubo()

        # Map each qubit to one or more symbols.
        num2syms = {}
        for s, n in self.sym_map.symbol_number_items():
            try:
                # Physical problem
                for pn in qprob.embedding[n]:
                    try:
                        num2syms[pn].append(s)
                    except KeyError:
                        num2syms[pn] = [s]
            except AttributeError:
                # Logical problem
                try:
                    num2syms[n].append(s)
                except KeyError:
                    num2syms[n] = [s]
        for n in num2syms.keys():
            num2syms[n].sort(key=lambda s: ("$" in s, s))

        # Find the character width of the longest list of symbol names.
        max_sym_name_len = max([len(repr(ss)) - 1 for ss in num2syms.values()] + [7])

        # Output all QMASM variables as MiniZinc variables.
        qubits_used = set(qprob.weights.keys())
        qubits_used.update([qs[0] for qs in qprob.strengths.keys()])
        qubits_used.update([qs[1] for qs in qprob.strengths.keys()])
        for q in sorted(qubits_used):
            outfile.write("var 0..1: q%d;  %% %s\n" % (q, " ".join(num2syms[q])))
        outfile.write("\n")

        # Define variables representing products of QMASM variables.  Constrain
        # the product variables to be the products.
        outfile.write("% Define p_X_Y variables and constrain them to be the product of qX and qY.\n")
        for q0, q1 in sorted(qprob.strengths.keys()):
            pstr = "p_%d_%d" % (q0, q1)
            outfile.write("var 0..1: %s;\n" % pstr)
            outfile.write("constraint %s >= q%d + q%d - 1;\n" % (pstr, q0, q1))
            outfile.write("constraint %s <= q%d;\n" % (pstr, q0))
            outfile.write("constraint %s <= q%d;\n" % (pstr, q1))
        outfile.write("\n")

        # Express energy as one, big Hamiltonian.
        scale_to_int = lambda f: int(round(10000.0*f))
        outfile.write("var int: energy =\n")
        weight_terms = ["%8d * q%d" % (scale_to_int(wt), q) for q, wt in sorted(qprob.weights.items())]
        strength_terms = ["%8d * p_%d_%d" % (scale_to_int(s), qs[0], qs[1]) for qs, s in sorted(qprob.strengths.items())]
        all_terms = weight_terms + strength_terms
        outfile.write("  %s;\n" % " +\n  ".join(all_terms))

        # Because we can't both minimize and enumerate all solutions, we
        # normally do only the former with instructions for the user on how to
        # switch to the latter.  However, if an energy was specified, comment
        # out the minimization step and uncomment the enumeration step.
        outfile.write("\n")
        outfile.write("% First pass: Compute the minimum energy.\n")
        if energy == None:
            outfile.write("solve minimize energy;\n")
        else:
            outfile.write("% solve minimize energy;\n")
        outfile.write("""
%% Second pass: Find all minimum-energy solutions.
%%
%% Once you've solved for minimum energy, comment out the "solve minimize
%% energy" line, plug the minimal energy value into the following line,
%% uncomment it and the "solve satisfy" line, and re-run MiniZinc, requesting
%% all solutions this time.  The catch is that you need to use the raw
%% energy value so be sure to modify the output block to show(energy)
%% instead of show(energy/%.10g + %.10g).
""" % (self.minizinc_scale_factor, qprob.bqm.offset))
        if energy == None:
            outfile.write("%constraint energy = -12345;\n")
            outfile.write("%solve satisfy;\n\n")
        else:
            outfile.write("constraint energy = %d;\n" % energy)
            outfile.write("solve satisfy;\n\n")

        # Output code to show the results symbolically.  We output in the same
        # format as QMASM normally does.  Unfortunately, I don't know how to get
        # MiniZinc to output the current solution number explicitly so I had to
        # hard-wire "Solution #1".
        outfile.write("output [\n")
        outfile.write('  "Solution #1 (energy = ", show(energy/%.10g + %.10g), ", tally = 1)\\n\\n",\n' % (self.minizinc_scale_factor, qprob.bqm.offset))
        outfile.write('  "    %-*s  Spin  Boolean\\n",\n' % (max_sym_name_len, "Name(s)"))
        outfile.write('  "    %s  ----  -------\\n",\n' % ("-" * max_sym_name_len))
        outlist = []
        for n, ss in num2syms.items():
            if ss == []:
                continue
            syms = " ".join(ss)
            line = ""
            line += '"    %-*s  ", ' % (max_sym_name_len, syms)
            if problem.qubo:
                line += 'show_int(4, q%d), ' % n
            else:
                line += 'show_int(4, 2*q%d - 1), ' % n
            line += '"  ", if show(q%d) == "1" then "True" else "False" endif, ' % n
            line += '"\\n"'
            outlist.append(line)
        outlist.sort()
        outfile.write("  %s\n];\n" % ",\n  ".join(outlist))

    def output_qbsolv(self, outfile, problem):
        "Output weights and strengths in qbsolv format."
        # Determine the list of nonzero weights and strengths.
        qprob = problem.convert_to_qubo()
        output_weights, output_strengths = qprob.weights, qprob.strengths
        max_node = max(list(output_weights.keys()) + [max(qs) for qs in output_strengths.keys()])
        num_nonzero_weights = len([q for q, wt in output_weights.items() if wt != 0.0])
        num_nonzero_strengths = len([qs for qs, wt in output_strengths.items() if wt != 0.0])

        # Assign dummy qubit numbers to qubits whose value is known a priori.
        try:
            n_known = len(qprob.known_values)
        except TypeError:
            n_known = 0
        try:
            extra_nodes = dict(zip(sorted(qprob.known_values.keys()),
                                   range(max_node + 1, max_node + 1 + n_known)))
        except AttributeError:
            extra_nodes = {}
        max_node += n_known
        num_nonzero_weights += n_known
        output_weights.update({num: qprob.known_values[sym]*qmasm.pin_weight
                               for sym, num in extra_nodes.items()})
        sym2num = dict(self.sym_map.symbol_number_items())
        sym2num.update(extra_nodes)

        # Output a name-to-number map as header comments.
        key_width = 0
        val_width = 0
        items = []
        for s, n in sym2num.items():
            if len(s) > key_width:
                key_width = len(s)

            # Map logical to physical if possible.
            try:
                # Physical problem
                known_values = qprob.logical.merged_known_values()
                pin_map = {k: v for k, v in qprob.logical.pinned}
            except AttributeError:
                # Logical problem
                known_values = qprob.merged_known_values()
                pin_map = {k: v for k, v in qprob.pinned}
            try:
                nstr = " ".join([str(n) for n in sorted(qprob.embedding[n])])
            except AttributeError:
                # Logical problem
                nstr = str(n)
            except KeyError:
                try:
                    nstr = "[Pinned to %s]" % repr(pin_map[n])
                except KeyError:
                    try:
                        nstr = "[Provably %s]" % known_values[n]
                    except KeyError:
                        try:
                            same = qprob.logical.contractions[n]
                            nstr = "[Same as %s]" % " ".join(self.sym_map.to_symbols(same))
                        except KeyError:
                            nstr = "[Disconnected]"
            if len(nstr) > val_width:
                val_width = len(nstr)
            items.append((s, nstr))
        items.sort()
        for s, nstr in items:
            outfile.write("c %-*s --> %-*s\n" % (key_width, s, val_width, nstr))

        # Output all nonzero weights and strengths.
        outfile.write("p qubo 0 %d %d %d\n" % (max_node + 1, num_nonzero_weights, num_nonzero_strengths))
        for q, wt in sorted(output_weights.items()):
            if wt != 0.0:
                outfile.write("%d %d %.10g\n" % (q, q, wt))
        for qs, wt in sorted(output_strengths.items()):
            if wt != 0.0:
                outfile.write("%d %d %.10g\n" % (qs[0], qs[1], wt))

    def output_qubist(self, outfile, as_qubo, problem, sampler):
        "Output weights and strengths in Qubist format, either Ising or QUBO."
        # Convert the problem to Ising, scale it for the hardware, then convert
        # to QUBO if requested.
        prob = problem.convert_to_ising()
        prob.autoscale_coefficients(sampler)
        if as_qubo:
            prob = prob.convert_to_qubo()
        output_weights = prob.weights
        output_strengths = prob.strengths

        # Format all weights and all strengths in Qubist format.
        data = []
        for q, wt in sorted(output_weights.items()):
            if wt != 0.0:
                data.append("%d %d %.10g" % (q, q, wt))
        for sp, str in sorted(output_strengths.items()):
            if str != 0.0:
                sp = sorted(sp)
                data.append("%d %d %.10g" % (sp[0], sp[1], str))

        # Output the header and data in Qubist format.
        try:
            num_qubits = sampler.sampler.properties["num_qubits"]
        except KeyError:
            # If the solver lacks a fixed hardware representation we assert
            # that the number of hardware qubits is exactly the number of
            # qubits we require.
            num_qubits = len(output_weights)
        outfile.write("%d %d\n" % (num_qubits, len(data)))
        for d in data:
            outfile.write("%s\n" % d)

    def write_output(self, problem, oname, oformat, as_qubo, sampler):
        "Write an output file in one of a variety of formats."

        # Open the output file.
        outfile = self.open_output_file(oname)

        # Output the weights and strengths in the specified format.
        if oformat == "qubist":
            self.output_qubist(outfile, as_qubo, problem, sampler)
        elif oformat == "qbsolv":
            self.output_qbsolv(outfile, problem)
        elif oformat == "qmasm":
            self.output_qmasm(outfile)
        elif oformat == "minizinc":
            self.output_minizinc(outfile, problem)
        elif oformat == "bqpjson":
            self.output_bqpjson(outfile, as_qubo, problem)

        # Close the output file.
        if oname != "<stdout>":
            outfile.close()



#! /usr/bin/env python

###################################
# Install QMASM                   #
# By Scott Pakin <pakin@lanl.gov> #
###################################

from setuptools import setup, find_packages

long_description = '''QMASM fills a gap in the software ecosystem for [D-Wave's adiabatic quantum computers](http://www.dwavesys.com/) by shielding the programmer from having to know system-specific hardware details while still enabling programs to be expressed at a fairly low level of abstraction.  It is therefore analogous to a conventional macro assembler and can be used in much the same way: as a target either for programmers who want a great deal of control over the hardware or for compilers that implement higher-level languages.'''

setup(name = 'qmasm',
      version = '4.0.1',
      description = 'Quantum Macro Assembler',
      long_description = long_description,
      long_description_content_type = 'text/markdown',
      author = 'Scott Pakin',
      author_email = 'pakin@lanl.gov',
      classifiers = [
          'Topic :: Software Development :: Compilers',
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Programming Language :: Python :: 3',
          'Intended Audience :: Developers'],
      url = 'https://github.com/lanl/qmasm',
      download_url = 'https://github.com/lanl/qmasm/archive/v4.0.tar.gz',
      license = 'BSD-ish',
      keywords = ['quantum', 'annealing', 'macro', 'assembler', 'd-wave'],
      entry_points = {
          'console_scripts': ['qmasm = qmasm.__main__:main']
      },
      scripts = ['scripts/qb2qmasm',
                 'extras/qmasm-gen-all-to-all',
                 'extras/qmasm-gen-chimera',
                 'extras/qmasm-gen-current'],
      packages = find_packages('src'),
      package_dir = {'': 'src'},
      python_requires = '>= 3.8',
      install_requires = ['dwave-ocean-sdk']
)



###################################
# QMASM utility functions         #
# By Scott Pakin <pakin@lanl.gov> #
###################################

from collections import defaultdict
import copy
import math
import qmasm
import sys

class RemainingNextException(Exception):
    'This exception is thrown if a "!next." directive can\'t be replaced.'
    pass

class Utilities(object):
    "Provide various utility functions as mixins for QMASM."

    def abend(self, str):
        "Abort the program on an error."
        sys.stderr.write("%s: %s\n" % (sys.argv[0], str))
        sys.exit(1)

    def warn(self, str):
        "Issue a warning message but continue execution."
        sys.stderr.write("%s: Warning: %s\n" % (sys.argv[0], str))

    def symbol_to_number(self, sym, prefix=None, next_prefix=None):
        "Map from a symbol to a number, creating a new association if necessary."
        # Replace "!next." by substituting prefixes in the name.
        if "!next." in sym:
            if prefix == None or next_prefix == None:
                raise RemainingNextException
            sym = sym.replace(prefix + "!next.", next_prefix)

        # Return the symbol's logical qubit number or allocate a new one.
        try:
            return self.sym_map.to_number(sym)
        except KeyError:
            return self.sym_map.new_symbol(sym)

    def apply_prefix(self, sym, prefix=None, next_prefix=None):
        "Apply a prefix to a symbol name, replacing !next. with the next prefix."
        if prefix != None:
            sym = prefix + sym
        if "!next." in sym:
            if prefix == None or next_prefix == None:
                raise RemainingNextException
            sym = sym.replace(prefix + "!next.", next_prefix)
        return sym

    def canonicalize_strengths(self, strs):
        "Combine edges (A, B) and (B, A) into (A, B) with A < B."
        new_strs = defaultdict(lambda: 0.0)
        for (q1, q2), wt in strs.items():
            if q1 == q2:
                continue          # Discard vertex weights.
            if wt == 0.0:
                continue          # Discard zero weights.
            if q1 > q2:
                q1, q2 = q2, q1   # Canonicalize vertex order.
            new_strs[(q1, q2)] += wt
        return new_strs

class SymbolMapping:
    "Map between symbols and numbers."

    def __init__(self):
        self.sym2num = {}
        self.num2syms = {}
        self.next_sym_num = 0

    def new_symbol(self, sym):
        "Assign the next available number to a symbol."
        if sym in self.sym2num:
            raise Exception("Internal error: Symbol %s is already defined" % sym)
        self.sym2num[sym] = self.next_sym_num
        self.num2syms[self.next_sym_num] = set([sym])
        self.next_sym_num += 1
        return self.sym2num[sym]

    def to_number(self, sym):
        "Map a symbol to a single number."
        return self.sym2num[sym]

    def to_symbols(self, num):
        "Map a number to a set of one or more symbols."
        return self.num2syms[num]

    def max_number(self):
        "Return the maximum symbol number used so far."
        return self.next_sym_num - 1

    def all_numbers(self):
        "Return an unordered list of all numbers used."
        return self.num2syms.keys()

    def all_symbols(self):
        "Return an unordered list of all symbols used."
        return self.sym2num.keys()

    def symbol_number_items(self):
        "Return a list of {symbol, number} pairs."
        return self.sym2num.items()

    def alias(self, sym1, sym2):
        "Make two symbols point to the same number."
        # Ensure that both symbols are defined.
        n_new_defs = 0   # Number of new definitions made
        try:
            num1 = self.sym2num[sym1]
        except KeyError:
            num1 = self.new_symbol(sym1)
            n_new_defs += 1
        try:
            num2 = self.sym2num[sym2]
        except KeyError:
            num2 = self.new_symbol(sym2)
            n_new_defs += 1

        # Do nothing if the symbols are already aliased.
        if num1 == num2:
            return num1

        # Abort if both symbols were previously defined (and do not alias each
        # other).  In this case, we'd need to merge the associated weights and
        # strengths, and we don't currently do that.
        if n_new_defs == 0:
            abend("Unable to alias pre-existing variables %s and %s" % (sym1, sym2))

        # Replace all occurrences of the larger number with the smaller in
        # num2syms.
        new_num = min(num1, num2)
        old_num = max(num1, num2)
        self.num2syms[new_num].update(self.num2syms[old_num])
        del self.num2syms[old_num]

        # Regenerate sym2num from num2syms.
        self.sym2num = {}
        for n, ss in self.num2syms.items():
            for s in ss:
                self.sym2num[s] = n
        return new_num

    def overwrite_with(self, sym2num):
        "Overwrite the map's contents with a given symbol-to-number map."
        self.sym2num = sym2num
        self.num2syms = {}
        for s, n in self.sym2num.items():
            try:
                self.num2syms[n].add(s)
            except KeyError:
                self.num2syms[n] = set([s])
        if len(self.num2syms.keys()) == 0:
            self.next_sym_num = 0
        else:
            self.next_sym_num = max(self.num2syms.keys()) + 1

    def replace_all(self, before_syms, after_syms):
        "Replace each symbol in before_syms with its peer in after_syms."
        sym2num = copy.deepcopy(self.sym2num)
        before_nums = []
        for s in before_syms:
            try:
                before_nums.append(sym2num[s])
            except KeyError:
                abend('Failed to rename nonexistent symbol "%s"' % s)
            del sym2num[s]
        for s, n in zip(after_syms, before_nums):
            sym2num[s] = n
        self.overwrite_with(sym2num)
